from __future__ import annotations

import itertools
import math
import random
from collections.abc import Iterable
from typing import (
    Any,
    Protocol,
    SupportsFloat,
    SupportsIndex,
    TypeVar,
    Union,
)

from s2clientprotocol import common_pb2 as common_pb


class HasPosition2D(Protocol):
    @property
    def position(self) -> Point2: ...


_PointLike = Union[tuple[float, float], tuple[float, float], tuple[float, ...]]
_PosLike = Union[HasPosition2D, _PointLike]
_TPosLike = TypeVar("_TPosLike", bound=_PosLike)

EPSILON: float = 10**-8


def _sign(num: SupportsFloat | SupportsIndex) -> float:
    return math.copysign(1, num)


class Pointlike(tuple[float, ...]):
    T = TypeVar("T", bound="Pointlike")

    @property
    def position(self: T) -> T:
        return self

    def distance_to(self, target: _PosLike) -> float:
        """Calculate a single distance from a point or unit to another point or unit

        :param target:"""
        p: tuple[float, ...] = target if isinstance(target, tuple) else target.position
        return math.hypot(self[0] - p[0], self[1] - p[1])

    def distance_to_point2(self, p: _PointLike) -> float:
        """Same as the function above, but should be a bit faster because of the dropped asserts
        and conversion.

        :param p:"""
        return math.hypot(self[0] - p[0], self[1] - p[1])

    def _distance_squared(self, p2: _PointLike) -> float:
        """Function used to not take the square root as the distances will stay proportionally the same.
        This is to speed up the sorting process.

        :param p2:"""
        return (self[0] - p2[0]) ** 2 + (self[1] - p2[1]) ** 2

    def sort_by_distance(self, ps: Iterable[_TPosLike]) -> list[_TPosLike]:
        """This returns the target points sorted as list.
        You should not pass a set or dict since those are not sortable.
        If you want to sort your units towards a point, use 'units.sorted_by_distance_to(point)' instead.

        :param ps:"""
        return sorted(ps, key=lambda p: self.distance_to_point2(p if isinstance(p, tuple) else p.position))

    def closest(self, ps: Iterable[_TPosLike]) -> _TPosLike:
        """This function assumes the 2d distance is meant

        :param ps:"""
        assert ps, "ps is empty"

        return min(ps, key=lambda p: self.distance_to_point2(p if isinstance(p, tuple) else p.position))

    def distance_to_closest(self, ps: Iterable[_TPosLike]) -> float:
        """This function assumes the 2d distance is meant
        :param ps:"""
        assert ps, "ps is empty"
        closest_distance = math.inf
        for p in ps:
            p2: tuple[float, ...] = p if isinstance(p, tuple) else p.position
            distance = self.distance_to_point2(p2)
            if distance <= closest_distance:
                closest_distance = distance
        return closest_distance

    def furthest(self, ps: Iterable[_TPosLike]) -> _TPosLike:
        """This function assumes the 2d distance is meant

        :param ps: Units object, or iterable of Unit or Point2"""
        assert ps, "ps is empty"

        return max(ps, key=lambda p: self.distance_to_point2(p if isinstance(p, tuple) else p.position))

    def distance_to_furthest(self, ps: Iterable[_PosLike]) -> float:
        """This function assumes the 2d distance is meant

        :param ps:"""
        assert ps, "ps is empty"
        furthest_distance = -math.inf
        for p in ps:
            p2: tuple[float, ...] = p if isinstance(p, tuple) else p.position
            distance = self.distance_to_point2(p2)
            if distance >= furthest_distance:
                furthest_distance = distance
        return furthest_distance

    def offset(self: T, p: _PointLike) -> T:
        """

        :param p:
        """
        return self.__class__(a + b for a, b in itertools.zip_longest(self, p[: len(self)], fillvalue=0))

    def unit_axes_towards(self: T, p: _PointLike) -> T:
        """

        :param p:
        """
        return self.__class__(_sign(b - a) for a, b in itertools.zip_longest(self, p[: len(self)], fillvalue=0))

    def towards(self: T, p: _PosLike, distance: float = 1, limit: bool = False) -> T:
        """

        :param p:
        :param distance:
        :param limit:
        """
        p2: tuple[float, ...] = p if isinstance(p, tuple) else p.position
        # assert self != p, f"self is {self}, p is {p}"
        # TODO test and fix this if statement
        if self == p2:
            return self
        # end of test
        d = self.distance_to_point2(p2)
        if limit:
            distance = min(d, distance)
        return self.__class__(
            a + (b - a) / d * distance for a, b in itertools.zip_longest(self, p2[: len(self)], fillvalue=0)
        )

    def __eq__(self, other: Any) -> bool:
        try:
            return all(abs(a - b) <= EPSILON for a, b in itertools.zip_longest(self, other, fillvalue=0))
        except TypeError:
            return False

    def __hash__(self) -> int:
        return hash(tuple(self))


class Point2(Pointlike):
    T = TypeVar("T", bound="Point2")

    @classmethod
    def from_proto(
        cls, data: common_pb.Point | common_pb.Point2D | common_pb.Size2DI | common_pb.PointI | Point2 | Point3
    ) -> Point2:
        """
        :param data:
        """
        return cls((data.x, data.y))

    @property
    def as_Point2D(self) -> common_pb.Point2D:
        return common_pb.Point2D(x=self.x, y=self.y)

    @property
    def as_PointI(self) -> common_pb.PointI:
        """Represents points on the minimap. Values must be between 0 and 64."""
        return common_pb.PointI(x=int(self[0]), y=int(self[1]))

    @property
    def rounded(self) -> Point2:
        return Point2((math.floor(self[0]), math.floor(self[1])))

    @property
    def length(self) -> float:
        """This property exists in case Point2 is used as a vector."""
        return math.hypot(self[0], self[1])

    @property
    def normalized(self: Point2 | Point3) -> Point2:
        """This property exists in case Point2 is used as a vector."""
        length = self.length
        # Cannot normalize if length is zero
        assert length
        return Point2((self[0] / length, self[1] / length))

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def to2(self) -> Point2:
        return Point2(self[:2])

    @property
    def to3(self) -> Point3:
        return Point3((*self, 0))

    def round(self, decimals: int) -> Point2:
        """Rounds each number in the tuple to the amount of given decimals."""
        return Point2((round(self[0], decimals), round(self[1], decimals)))

    def offset(self: T, p: _PointLike) -> T:
        return self.__class__((self[0] + p[0], self[1] + p[1]))

    def random_on_distance(self, distance: float | tuple[float, float] | list[float]) -> Point2:
        if isinstance(distance, (tuple, list)):  # interval
            dist = distance[0] + random.random() * (distance[1] - distance[0])
        else:
            dist = distance
        assert dist > 0, "Distance is not greater than 0"
        angle = random.random() * 2 * math.pi

        dx, dy = math.cos(angle), math.sin(angle)
        return Point2((self.x + dx * dist, self.y + dy * dist))

    def towards_with_random_angle(
        self,
        p: Point2 | Point3,
        distance: int | float = 1,
        max_difference: int | float = (math.pi / 4),
    ) -> Point2:
        tx, ty = self.to2.towards(p.to2, 1)
        angle = math.atan2(ty - self.y, tx - self.x)
        angle = (angle - max_difference) + max_difference * 2 * random.random()
        return Point2((self.x + math.cos(angle) * distance, self.y + math.sin(angle) * distance))

    def circle_intersection(self, p: Point2, r: float) -> set[Point2]:
        """self is point1, p is point2, r is the radius for circles originating in both points
        Used in ramp finding

        :param p:
        :param r:"""
        assert self != p, "self is equal to p"
        distance_between_points = self.distance_to(p)
        assert r >= distance_between_points / 2
        # remaining distance from center towards the intersection, using pythagoras
        remaining_distance_from_center = (r**2 - (distance_between_points / 2) ** 2) ** 0.5
        # center of both points
        offset_to_center = Point2(((p.x - self.x) / 2, (p.y - self.y) / 2))
        center = self.offset(offset_to_center)

        # stretch offset vector in the ratio of remaining distance from center to intersection
        vector_stretch_factor = remaining_distance_from_center / (distance_between_points / 2)
        v = offset_to_center
        offset_to_center_stretched = Point2((v.x * vector_stretch_factor, v.y * vector_stretch_factor))

        # rotate vector by 90° and -90°
        vector_rotated_1 = Point2((offset_to_center_stretched.y, -offset_to_center_stretched.x))
        vector_rotated_2 = Point2((-offset_to_center_stretched.y, offset_to_center_stretched.x))
        intersect1 = center.offset(vector_rotated_1)
        intersect2 = center.offset(vector_rotated_2)
        return {intersect1, intersect2}

    @property
    def neighbors4(self: T) -> set[T]:
        return {
            self.__class__((self[0] - 1, self[1])),
            self.__class__((self[0] + 1, self[1])),
            self.__class__((self[0], self[1] - 1)),
            self.__class__((self[0], self[1] + 1)),
        }

    @property
    def neighbors8(self: T) -> set[T]:
        return self.neighbors4 | {
            self.__class__((self[0] - 1, self[1] - 1)),
            self.__class__((self[0] - 1, self[1] + 1)),
            self.__class__((self[0] + 1, self[1] - 1)),
            self.__class__((self[0] + 1, self[1] + 1)),
        }

    def negative_offset(self: T, other: Point2) -> T:
        return self.__class__((self[0] - other[0], self[1] - other[1]))

    def __add__(self, other: Point2) -> Point2:  # pyright: ignore[reportIncompatibleMethodOverride]
        return self.offset(other)

    def __sub__(self, other: Point2) -> Point2:
        return self.negative_offset(other)

    def __neg__(self: T) -> T:
        return self.__class__(-a for a in self)

    def __abs__(self) -> float:
        return math.hypot(self[0], self[1])

    def __bool__(self) -> bool:
        return self[0] != 0 or self[1] != 0

    def __mul__(self, other: _PointLike | float) -> Point2:  # pyright: ignore[reportIncompatibleMethodOverride]
        if isinstance(other, (int, float)):
            return Point2((self[0] * other, self[1] * other))
        return Point2((self[0] * other[0], self[1] * other[1]))

    def __rmul__(self, other: _PointLike | float) -> Point2:  # pyright: ignore[reportIncompatibleMethodOverride]
        return self.__mul__(other)

    def __truediv__(self, other: float | Point2) -> Point2:
        if isinstance(other, (int, float)):
            return self.__class__((self[0] / other, self[1] / other))
        return self.__class__((self[0] / other[0], self[1] / other[1]))

    def is_same_as(self, other: Point2, dist: float = 0.001) -> bool:
        return self.distance_to_point2(other) <= dist

    def direction_vector(self, other: Point2) -> Point2:
        """Converts a vector to a direction that can face vertically, horizontally or diagonal or be zero, e.g. (0, 0), (1, -1), (1, 0)"""
        return self.__class__((_sign(other[0] - self[0]), _sign(other[1] - self[1])))

    def manhattan_distance(self, other: Point2) -> float:
        """
        :param other:
        """
        return abs(other[0] - self[0]) + abs(other[1] - self[1])

    @staticmethod
    def center(points: list[Point2]) -> Point2:
        """Returns the central point for points in list

        :param points:"""
        s = Point2((0, 0))
        for p in points:
            s += p
        return s / len(points)


class Point3(Point2):
    @classmethod
    def from_proto(cls, data: common_pb.Point | Point3) -> Point3:  # pyright: ignore[reportIncompatibleMethodOverride]
        """
        :param data:
        """
        return cls((data.x, data.y, data.z))

    @property
    def as_Point(self) -> common_pb.Point:
        return common_pb.Point(x=self.x, y=self.y, z=self.z)

    @property
    def rounded(self) -> Point3:
        return Point3((math.floor(self[0]), math.floor(self[1]), math.floor(self[2])))

    @property
    def z(self) -> float:
        return self[2]

    @property
    def to3(self) -> Point3:
        return Point3(self)

    def __add__(self, other: Point2 | Point3) -> Point3:
        if not isinstance(other, Point3):
            return Point3((self[0] + other[0], self[1] + other[1], self[2]))
        return Point3((self[0] + other[0], self[1] + other[1], self[2] + other[2]))


class Size(Point2):
    @classmethod
    def from_proto(
        cls, data: common_pb.Point | common_pb.Point2D | common_pb.Size2DI | common_pb.PointI | Point2
    ) -> Size:
        """
        :param data:
        """
        return cls((data.x, data.y))

    @property
    def width(self) -> float:
        return self[0]

    @property
    def height(self) -> float:
        return self[1]


class Rect(Point2):
    @classmethod
    def from_proto(cls, data: common_pb.RectangleI) -> Rect:  # pyright: ignore[reportIncompatibleMethodOverride]
        """
        :param data:
        """
        assert data.p0.x < data.p1.x and data.p0.y < data.p1.y
        return cls((data.p0.x, data.p0.y, data.p1.x - data.p0.x, data.p1.y - data.p0.y))

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def width(self) -> float:
        return self[2]

    @property
    def height(self) -> float:
        return self[3]

    @property
    def right(self) -> float:
        """Returns the x-coordinate of the rectangle of its right side."""
        return self.x + self.width

    @property
    def top(self) -> float:
        """Returns the y-coordinate of the rectangle of its top side."""
        return self.y + self.height

    @property
    def size(self) -> Size:
        return Size((self[2], self[3]))

    @property
    def center(self) -> Point2:  # pyright: ignore[reportIncompatibleMethodOverride]
        return Point2((self.x + self.width / 2, self.y + self.height / 2))

    def offset(self, p: _PointLike) -> Rect:
        return self.__class__((self[0] + p[0], self[1] + p[1], self[2], self[3]))



from __future__ import annotations

from collections.abc import Callable, Hashable
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from sc2.bot_ai import BotAI

T = TypeVar("T")


class CacheDict(dict[Hashable, Any]):
    def retrieve_and_set(self, key: Hashable, func: Callable[[], T]) -> T:
        """Either return the value at a certain key,
        or set the return value of a function to that key, then return that value."""
        if key not in self:
            self[key] = func()
        return self[key]


class property_cache_once_per_frame(property):  # noqa: N801
    """This decorator caches the return value for one game loop,
    then clears it if it is accessed in a different game loop.
    Only works on properties of the bot object, because it requires
    access to self.state.game_loop

    This decorator compared to the above runs a little faster, however you should only use this decorator if you are sure that you do not modify the mutable once it is calculated and cached.

    Copied and modified from https://tedboy.github.io/flask/_modules/werkzeug/utils.html#cached_property
    #"""

    def __init__(self, func: Callable[[BotAI], T], name: str | None = None) -> None:
        self.__name__ = name or func.__name__
        self.__frame__ = f"__frame__{self.__name__}"
        self.func = func

    def __set__(self, obj: BotAI, value: T) -> None:
        obj.cache[self.__name__] = value
        # pyre-ignore[16]
        obj.cache[self.__frame__] = obj.state.game_loop

    # pyre-fixme[34]
    def __get__(self, obj: BotAI, _type=None) -> T:
        value = obj.cache.get(self.__name__, None)
        # pyre-ignore[16]
        bot_frame = obj.state.game_loop
        if value is None or obj.cache[self.__frame__] < bot_frame:
            value = self.func(obj)
            obj.cache[self.__name__] = value
            obj.cache[self.__frame__] = bot_frame
        return value



from __future__ import annotations

import asyncio
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
from contextlib import suppress
from pathlib import Path
from typing import Any

import aiohttp
import portpicker
from aiohttp.client_ws import ClientWebSocketResponse
from loguru import logger

from sc2 import paths, wsl
from sc2.controller import Controller
from sc2.paths import Paths
from sc2.versions import VERSIONS


class KillSwitch:
    _to_kill: list[Any] = []

    @classmethod
    def add(cls, value) -> None:
        logger.debug("kill_switch: Add switch")
        cls._to_kill.append(value)

    @classmethod
    def kill_all(cls) -> None:
        logger.info(f"kill_switch: Process cleanup for {len(cls._to_kill)} processes")
        for p in cls._to_kill:
            p._clean(verbose=False)


class SC2Process:
    """
    A class for handling SCII applications.

    :param host: hostname for the url the SCII application will listen to
    :param port: the websocket port the SCII application will listen to
    :param fullscreen: whether to launch the SCII application in fullscreen or not, defaults to False
    :param resolution: (window width, window height) in pixels, defaults to (1024, 768)
    :param placement: (x, y) the distances of the SCII app's top left corner from the top left corner of the screen
                       e.g. (20, 30) is 20 to the right of the screen's left border, and 30 below the top border
    :param render:
    :param sc2_version:
    :param base_build:
    :param data_hash:
    """

    def __init__(
        self,
        host: str | None = None,
        port: int | None = None,
        fullscreen: bool = False,
        resolution: list[int] | tuple[int, int] | None = None,
        placement: list[int] | tuple[int, int] | None = None,
        render: bool = False,
        sc2_version: str | None = None,
        base_build: str | None = None,
        data_hash: str | None = None,
    ) -> None:
        assert isinstance(host, str) or host is None
        assert isinstance(port, int) or port is None

        self._render = render
        self._arguments: dict[str, str] = {"-displayMode": str(int(fullscreen))}
        if not fullscreen:
            if resolution and len(resolution) == 2:
                self._arguments["-windowwidth"] = str(resolution[0])
                self._arguments["-windowheight"] = str(resolution[1])
            if placement and len(placement) == 2:
                self._arguments["-windowx"] = str(placement[0])
                self._arguments["-windowy"] = str(placement[1])

        self._host = host or os.environ.get("SC2CLIENTHOST", "127.0.0.1")
        self._serverhost = os.environ.get("SC2SERVERHOST", self._host)

        if port is None:
            self._port = portpicker.pick_unused_port()
        else:
            self._port = port
        self._used_portpicker = bool(port is None)
        self._tmp_dir = tempfile.mkdtemp(prefix="SC2_")
        self._process: subprocess.Popen | None = None
        self._session = None
        self._ws = None
        self._sc2_version = sc2_version
        self._base_build = base_build
        self._data_hash = data_hash

    async def __aenter__(self) -> Controller:
        KillSwitch.add(self)

        def signal_handler(*_args):
            # unused arguments: signal handling library expects all signal
            # callback handlers to accept two positional arguments
            KillSwitch.kill_all()

        signal.signal(signal.SIGINT, signal_handler)

        try:
            self._process = self._launch()
            self._ws = await self._connect()
        except:
            await self._close_connection()
            self._clean()
            raise

        return Controller(self._ws, self)

    async def __aexit__(self, *args) -> None:
        await self._close_connection()
        KillSwitch.kill_all()
        signal.signal(signal.SIGINT, signal.SIG_DFL)

    @property
    def ws_url(self) -> str:
        return f"ws://{self._host}:{self._port}/sc2api"

    @property
    def versions(self):
        """Opens the versions.json file which origins from
        https://github.com/Blizzard/s2client-proto/blob/master/buildinfo/versions.json"""
        return VERSIONS

    def find_data_hash(self, target_sc2_version: str) -> str | None:
        """Returns the data hash from the matching version string."""
        version: dict
        for version in self.versions:
            if version["label"] == target_sc2_version:
                return version["data-hash"]
        return None

    def find_base_dir(self, target_sc2_version: str) -> str | None:
        """Returns the base directory from the matching version string."""
        for version in self.versions:
            if version["label"] == target_sc2_version:
                return "Base" + str(version["base-version"])
        return None

    def _launch(self):
        if self._sc2_version and not self._base_build:
            self._base_build = self.find_base_dir(self._sc2_version)

        if self._base_build:
            executable = str(paths.latest_executeble(Paths.BASE / "Versions", self._base_build))
        else:
            executable = str(Paths.EXECUTABLE)

        if self._port is None:
            self._port = portpicker.pick_unused_port()
            self._used_portpicker = True
        args = paths.get_runner_args(Paths.CWD) + [
            executable,
            "-listen",
            self._serverhost,
            "-port",
            str(self._port),
            "-dataDir",
            str(Paths.BASE),
            "-tempDir",
            self._tmp_dir,
        ]
        for arg, value in self._arguments.items():
            args.append(arg)
            args.append(value)
        if self._sc2_version:

            def special_match(strg: str):
                """Tests if the specified version is in the versions.py dict."""
                return any(version["label"] == strg for version in self.versions)

            valid_version_string = special_match(self._sc2_version)
            if valid_version_string:
                self._data_hash = self.find_data_hash(self._sc2_version)
                assert self._data_hash is not None, (
                    f"StarCraft 2 Client version ({self._sc2_version}) was not found inside sc2/versions.py file. Please check your spelling or check the versions.py file."
                )

            else:
                logger.warning(
                    f'The submitted version string in sc2.rungame() function call (sc2_version="{self._sc2_version}") was not found in versions.py. Running latest version instead.'
                )

        if self._data_hash:
            args.extend(["-dataVersion", self._data_hash])

        if self._render:
            args.extend(["-eglpath", "libEGL.so"])

        # if logger.getEffectiveLevel() <= logging.DEBUG:
        args.append("-verbose")

        sc2_cwd = str(Paths.CWD) if Paths.CWD else None

        if paths.PF in {"WSL1", "WSL2"}:
            return wsl.run(args, sc2_cwd)

        return subprocess.Popen(
            args,
            cwd=sc2_cwd,
            # Suppress Wine error messages
            stderr=subprocess.DEVNULL,
            # , env=run_config.env
        )

    async def _connect(self) -> ClientWebSocketResponse:
        # How long it waits for SC2 to start (in seconds)
        for i in range(180):
            if self._process is None:
                # The ._clean() was called, clearing the process
                logger.debug("Process cleanup complete, exit")
                sys.exit()

            await asyncio.sleep(1)
            try:
                self._session = aiohttp.ClientSession()
                ws = await self._session.ws_connect(self.ws_url, timeout=120)
                # FIXME fix deprecation warning in for future aiohttp version
                # ws = await self._session.ws_connect(
                #     self.ws_url, timeout=aiohttp.client_ws.ClientWSTimeout(ws_close=120)
                # )
                logger.debug("Websocket connection ready")
                return ws
            except aiohttp.client_exceptions.ClientConnectorError:
                await self._session.close()
                if i > 15:
                    logger.debug("Connection refused (startup not complete (yet))")

        logger.debug("Websocket connection to SC2 process timed out")
        raise TimeoutError("Websocket")

    async def _close_connection(self) -> None:
        logger.info(f"Closing connection at {self._port}...")

        if self._ws is not None:
            await self._ws.close()

        if self._session is not None:
            await self._session.close()

    def _clean(self, verbose: bool = True) -> None:
        if verbose:
            logger.info("Cleaning up...")

        if self._process is not None:
            assert isinstance(self._process, subprocess.Popen)
            if paths.PF in {"WSL1", "WSL2"}:
                if wsl.kill(self._process):
                    logger.error("KILLED")
            elif self._process.poll() is None:
                for _ in range(3):
                    self._process.terminate()
                    time.sleep(0.5)
                    if not self._process or self._process.poll() is not None:
                        break
            else:
                self._process.kill()
                self._process.wait()
                logger.error("KILLED")
            # Try to kill wineserver on linux
            if paths.PF in {"Linux", "WineLinux"}:
                # Command wineserver not detected
                with suppress(FileNotFoundError), subprocess.Popen(["wineserver", "-k"]) as p:
                    p.wait()

        if Path(self._tmp_dir).exists():
            shutil.rmtree(self._tmp_dir)

        self._process = None
        self._ws = None
        if self._used_portpicker and self._port is not None:
            portpicker.return_port(self._port)
            self._port = None
        if verbose:
            logger.info("Cleanup complete")



# pyre-ignore-all-errors[6, 11, 16, 58]
from __future__ import annotations

import heapq
from collections import deque
from collections.abc import Iterable
from dataclasses import dataclass
from functools import cached_property

import numpy as np

from s2clientprotocol import sc2api_pb2
from sc2.pixel_map import PixelMap
from sc2.player import Player
from sc2.position import Point2, Rect, Size


@dataclass
class Ramp:
    points: frozenset[Point2]
    game_info: GameInfo

    @property
    def x_offset(self) -> float:
        # Tested by printing actual building locations vs calculated depot positions
        return 0.5

    @property
    def y_offset(self) -> float:
        # Tested by printing actual building locations vs calculated depot positions
        return 0.5

    @cached_property
    def _height_map(self):
        return self.game_info.terrain_height

    @cached_property
    def size(self) -> int:
        return len(self.points)

    def height_at(self, p: Point2) -> int:
        return self._height_map[p]

    @cached_property
    def upper(self) -> frozenset[Point2]:
        """Returns the upper points of a ramp."""
        current_max = -10000
        result = set()
        for p in self.points:
            height = self.height_at(p)
            if height > current_max:
                current_max = height
                result = {p}
            elif height == current_max:
                result.add(p)
        return frozenset(result)

    @cached_property
    def upper2_for_ramp_wall(self) -> frozenset[Point2]:
        """Returns the 2 upper ramp points of the main base ramp required for the supply depot and barracks placement properties used in this file."""
        # From bottom center, find 2 points that are furthest away (within the same ramp)
        return frozenset(heapq.nlargest(2, self.upper, key=lambda x: x.distance_to_point2(self.bottom_center)))

    @cached_property
    def top_center(self) -> Point2:
        length = len(self.upper)
        pos = Point2((sum(p.x for p in self.upper) / length, sum(p.y for p in self.upper) / length))
        return pos

    @cached_property
    def lower(self) -> frozenset[Point2]:
        current_min = 10000
        result = set()
        for p in self.points:
            height = self.height_at(p)
            if height < current_min:
                current_min = height
                result = {p}
            elif height == current_min:
                result.add(p)
        return frozenset(result)

    @cached_property
    def bottom_center(self) -> Point2:
        length = len(self.lower)
        pos = Point2((sum(p.x for p in self.lower) / length, sum(p.y for p in self.lower) / length))
        return pos

    @cached_property
    def barracks_in_middle(self) -> Point2 | None:
        """Barracks position in the middle of the 2 depots"""
        if len(self.upper) not in {2, 5}:
            return None
        if len(self.upper2_for_ramp_wall) == 2:
            points = set(self.upper2_for_ramp_wall)
            p1 = points.pop().offset((self.x_offset, self.y_offset))
            p2 = points.pop().offset((self.x_offset, self.y_offset))
            # Offset from top point to barracks center is (2, 1)
            intersects = p1.circle_intersection(p2, 5**0.5)
            any_lower_point = next(iter(self.lower))
            return max(intersects, key=lambda p: p.distance_to_point2(any_lower_point))

        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @cached_property
    def depot_in_middle(self) -> Point2 | None:
        """Depot in the middle of the 3 depots"""
        if len(self.upper) not in {2, 5}:
            return None
        if len(self.upper2_for_ramp_wall) == 2:
            points = set(self.upper2_for_ramp_wall)
            p1 = points.pop().offset((self.x_offset, self.y_offset))
            p2 = points.pop().offset((self.x_offset, self.y_offset))
            # Offset from top point to depot center is (1.5, 0.5)
            try:
                intersects = p1.circle_intersection(p2, 2.5**0.5)
            except AssertionError:
                # Returns None when no placement was found, this is the case on the map Honorgrounds LE with an exceptionally large main base ramp
                return None
            any_lower_point = next(iter(self.lower))
            return max(intersects, key=lambda p: p.distance_to_point2(any_lower_point))

        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @cached_property
    def corner_depots(self) -> set[Point2]:
        """Finds the 2 depot positions on the outside"""
        if not self.upper2_for_ramp_wall:
            return set()
        if len(self.upper2_for_ramp_wall) == 2:
            points = set(self.upper2_for_ramp_wall)
            p1 = points.pop().offset((self.x_offset, self.y_offset))
            p2 = points.pop().offset((self.x_offset, self.y_offset))
            center = p1.towards(p2, p1.distance_to_point2(p2) / 2)
            depot_position = self.depot_in_middle
            if depot_position is None:
                return set()
            # Offset from middle depot to corner depots is (2, 1)
            intersects = center.circle_intersection(depot_position, 5**0.5)
            return intersects

        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @cached_property
    def barracks_can_fit_addon(self) -> bool:
        """Test if a barracks can fit an addon at natural ramp"""
        # https://i.imgur.com/4b2cXHZ.png
        if len(self.upper2_for_ramp_wall) == 2:
            return self.barracks_in_middle.x + 1 > max(self.corner_depots, key=lambda depot: depot.x).x

        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @cached_property
    def barracks_correct_placement(self) -> Point2 | None:
        """Corrected placement so that an addon can fit"""
        if self.barracks_in_middle is None:
            return None
        if len(self.upper2_for_ramp_wall) == 2:
            if self.barracks_can_fit_addon:
                return self.barracks_in_middle
            return self.barracks_in_middle.offset((-2, 0))

        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @cached_property
    def protoss_wall_pylon(self) -> Point2 | None:
        """
        Pylon position that powers the two wall buildings and the warpin position.
        """
        if len(self.upper) not in {2, 5}:
            return None
        if len(self.upper2_for_ramp_wall) != 2:
            raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
        middle = self.depot_in_middle
        # direction up the ramp
        direction = self.barracks_in_middle.negative_offset(middle)
        # pyre-ignore[7]
        return middle + 6 * direction

    @cached_property
    def protoss_wall_buildings(self) -> frozenset[Point2]:
        """
        List of two positions for 3x3 buildings that form a wall with a spot for a one unit block.
        These buildings can be powered by a pylon on the protoss_wall_pylon position.
        """
        if len(self.upper) not in {2, 5}:
            return frozenset()
        if len(self.upper2_for_ramp_wall) == 2:
            middle = self.depot_in_middle
            # direction up the ramp
            direction = self.barracks_in_middle.negative_offset(middle)
            # sort depots based on distance to start to get wallin orientation
            sorted_depots = sorted(
                self.corner_depots, key=lambda depot: depot.distance_to(self.game_info.player_start_location)
            )
            wall1: Point2 = sorted_depots[1].offset(direction)
            wall2 = middle + direction + (middle - wall1) / 1.5
            return frozenset([wall1, wall2])

        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @cached_property
    def protoss_wall_warpin(self) -> Point2 | None:
        """
        Position for a unit to block the wall created by protoss_wall_buildings.
        Powered by protoss_wall_pylon.
        """
        if len(self.upper) not in {2, 5}:
            return None
        if len(self.upper2_for_ramp_wall) != 2:
            raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")
        middle = self.depot_in_middle
        # direction up the ramp
        direction = self.barracks_in_middle.negative_offset(middle)
        # sort depots based on distance to start to get wallin orientation
        sorted_depots = sorted(self.corner_depots, key=lambda x: x.distance_to(self.game_info.player_start_location))
        return sorted_depots[0].negative_offset(direction)


class GameInfo:
    def __init__(self, proto: sc2api_pb2.ResponseGameInfo) -> None:
        self._proto = proto
        self.players: list[Player] = [Player.from_proto(p) for p in self._proto.player_info]
        self.map_name: str = self._proto.map_name
        self.local_map_path: str = self._proto.local_map_path
        # pyre-ignore[8]
        self.map_size: Size = Size.from_proto(self._proto.start_raw.map_size)

        # self.pathing_grid[point]: if 0, point is not pathable, if 1, point is pathable
        self.pathing_grid: PixelMap = PixelMap(self._proto.start_raw.pathing_grid, in_bits=True)
        # self.terrain_height[point]: returns the height in range of 0 to 255 at that point
        self.terrain_height: PixelMap = PixelMap(self._proto.start_raw.terrain_height)
        # self.placement_grid[point]: if 0, point is not placeable, if 1, point is pathable
        self.placement_grid: PixelMap = PixelMap(self._proto.start_raw.placement_grid, in_bits=True)
        self.playable_area = Rect.from_proto(self._proto.start_raw.playable_area)
        self.map_center = self.playable_area.center
        # pyre-ignore[8]
        self.map_ramps: list[Ramp] = None  # Filled later by BotAI._prepare_first_step
        # pyre-ignore[8]
        self.vision_blockers: frozenset[Point2] = None  # Filled later by BotAI._prepare_first_step
        self.player_races: dict[int, int] = {
            p.player_id: p.race_actual or p.race_requested for p in self._proto.player_info
        }
        self.start_locations: list[Point2] = [
            Point2.from_proto(sl).round(decimals=1) for sl in self._proto.start_raw.start_locations
        ]
        # pyre-ignore[8]
        self.player_start_location: Point2 = None  # Filled later by BotAI._prepare_first_step

    def _find_ramps_and_vision_blockers(self) -> tuple[list[Ramp], frozenset[Point2]]:
        """Calculate points that are pathable but not placeable.
        Then divide them into ramp points if not all points around the points are equal height
        and into vision blockers if they are."""

        def equal_height_around(tile):
            # mask to slice array 1 around tile
            sliced = self.terrain_height.data_numpy[tile[1] - 1 : tile[1] + 2, tile[0] - 1 : tile[0] + 2]
            return len(np.unique(sliced)) == 1

        map_area = self.playable_area
        # all points in the playable area that are pathable but not placable
        points = [
            Point2((a, b))
            for (b, a), value in np.ndenumerate(self.pathing_grid.data_numpy)
            if value == 1
            and map_area.x <= a < map_area.x + map_area.width
            and map_area.y <= b < map_area.y + map_area.height
            and self.placement_grid[(a, b)] == 0
        ]
        # divide points into ramp points and vision blockers
        ramp_points = [point for point in points if not equal_height_around(point)]
        vision_blockers = frozenset(point for point in points if equal_height_around(point))
        ramps = [Ramp(group, self) for group in self._find_groups(ramp_points)]
        return ramps, vision_blockers

    def _find_groups(self, points: frozenset[Point2], minimum_points_per_group: int = 8) -> Iterable[frozenset[Point2]]:
        """
        From a set of points, this function will try to group points together by
        painting clusters of points in a rectangular map using flood fill algorithm.
        Returns groups of points as list, like [{p1, p2, p3}, {p4, p5, p6, p7, p8}]
        """
        # TODO do we actually need colors here? the ramps will never touch anyways.
        NOT_COLORED_YET = -1
        map_width = self.pathing_grid.width
        map_height = self.pathing_grid.height
        current_color: int = NOT_COLORED_YET
        picture: list[list[int]] = [[-2 for _ in range(map_width)] for _ in range(map_height)]

        def paint(pt: Point2) -> None:
            picture[pt.y][pt.x] = current_color

        nearby: list[tuple[int, int]] = [(a, b) for a in [-1, 0, 1] for b in [-1, 0, 1] if a != 0 or b != 0]

        remaining: set[Point2] = set(points)
        for point in remaining:
            paint(point)
        current_color = 1
        queue: deque[Point2] = deque()
        while remaining:
            current_group: set[Point2] = set()
            if not queue:
                start = remaining.pop()
                paint(start)
                queue.append(start)
                current_group.add(start)
            while queue:
                base: Point2 = queue.popleft()
                for offset in nearby:
                    px, py = base.x + offset[0], base.y + offset[1]
                    # Do we ever reach out of map bounds?
                    if not (0 <= px < map_width and 0 <= py < map_height):
                        continue
                    if picture[py][px] != NOT_COLORED_YET:
                        continue
                    point: Point2 = Point2((px, py))
                    remaining.discard(point)
                    paint(point)
                    queue.append(point)
                    current_group.add(point)
            if len(current_group) >= minimum_points_per_group:
                yield frozenset(current_group)



from collections.abc import Iterable
from enum import Enum

from google.protobuf.message import Message

from .common_pb2 import Point, Point2D

class DebugCommand(Message):
    draw: DebugDraw
    game_state: int
    create_unit: DebugCreateUnit
    kill_unit: DebugKillUnit
    test_process: DebugTestProcess
    score: DebugSetScore
    end_game: DebugEndGame
    unit_value: DebugSetUnitValue
    def __init__(
        self,
        draw: DebugDraw = ...,
        game_state: int = ...,
        create_unit: DebugCreateUnit = ...,
        kill_unit: DebugKillUnit = ...,
        test_process: DebugTestProcess = ...,
        score: DebugSetScore = ...,
        end_game: DebugEndGame = ...,
        unit_value: DebugSetUnitValue = ...,
    ) -> None: ...

class DebugDraw(Message):
    text: Iterable[DebugText]
    lines: Iterable[DebugLine]
    boxes: Iterable[DebugBox]
    spheres: Iterable[DebugSphere]
    def __init__(
        self,
        text: Iterable[DebugText] = ...,
        lines: Iterable[DebugLine] = ...,
        boxes: Iterable[DebugBox] = ...,
        spheres: Iterable[DebugSphere] = ...,
    ) -> None: ...

class Line(Message):
    p0: Point
    p1: Point
    def __init__(self, p0: Point = ..., p1: Point = ...) -> None: ...

class Color(Message):
    r: int
    g: int
    b: int
    def __init__(self, r: int = ..., g: int = ..., b: int = ...) -> None: ...

class DebugText(Message):
    color: Color
    text: str
    virtual_pos: Point
    world_pos: Point
    size: int
    def __init__(
        self,
        color: Color = ...,
        text: str = ...,
        virtual_pos: Point = ...,
        world_pos: Point = ...,
        size: int = ...,
    ) -> None: ...

class DebugLine(Message):
    color: Color
    line: Line
    def __init__(self, color: Color = ..., line: Line = ...) -> None: ...

class DebugBox(Message):
    color: Color
    min: Point
    max: Point
    def __init__(self, color: Color = ..., min: Point = ..., max: Point = ...) -> None: ...

class DebugSphere(Message):
    color: Color
    p: Point
    r: float
    def __init__(self, color: Color = ..., p: Point = ..., r: float = ...) -> None: ...

class DebugGameState(Enum):
    show_map: int
    control_enemy: int
    food: int
    free: int
    all_resources: int
    god: int
    minerals: int
    gas: int
    cooldown: int
    tech_tree: int
    upgrade: int
    fast_build: int

class DebugCreateUnit(Message):
    unit_type: int
    owner: int
    pos: Point2D
    quantity: int
    def __init__(
        self,
        unit_type: int = ...,
        owner: int = ...,
        pos: Point2D = ...,
        quantity: int = ...,
    ) -> None: ...

class DebugKillUnit(Message):
    tag: Iterable[int]
    def __init__(self, tag: Iterable[int] = ...) -> None: ...

class Test(Enum):
    hang: int
    crash: int
    exit: int

class DebugTestProcess(Message):
    test: int
    delay_ms: int
    def __init__(self, test: int = ..., delay_ms: int = ...) -> None: ...

class DebugSetScore(Message):
    score: float
    def __init__(self, score: float = ...) -> None: ...

class EndResult(Enum):
    Surrender: int
    DeclareVictory: int

class DebugEndGame(Message):
    end_result: int
    def __init__(self, end_result: int = ...) -> None: ...

class UnitValue(Enum):
    Energy: int
    Life: int
    Shields: int

class DebugSetUnitValue(Message):
    unit_value: int
    value: float
    unit_tag: int
    def __init__(
        self,
        unit_value: int = ...,
        value: float = ...,
        unit_tag: int = ...,
    ) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING

from sc2.constants import COMBINEABLE_ABILITIES
from sc2.ids.ability_id import AbilityId
from sc2.position import Point2

if TYPE_CHECKING:
    from sc2.unit import Unit


class UnitCommand:
    def __init__(
        self, ability: AbilityId, unit: Unit, target: Unit | Point2 | None = None, queue: bool = False
    ) -> None:
        """
        :param ability:
        :param unit:
        :param target:
        :param queue:
        """
        assert ability in AbilityId, f"ability {ability} is not in AbilityId"
        assert unit.__class__.__name__ == "Unit", f"unit {unit} is of type {type(unit)}"
        assert any(
            [
                target is None,
                isinstance(target, Point2),
                unit.__class__.__name__ == "Unit",
            ]
        ), f"target {target} is of type {type(target)}"
        assert isinstance(queue, bool), f"queue flag {queue} is of type {type(queue)}"
        self.ability = ability
        self.unit = unit
        self.target = target
        self.queue = queue

    @property
    def combining_tuple(self) -> tuple[AbilityId, Unit | Point2 | None, bool, bool]:
        return self.ability, self.target, self.queue, self.ability in COMBINEABLE_ABILITIES

    def __repr__(self) -> str:
        return f"UnitCommand({self.ability}, {self.unit}, {self.target}, {self.queue})"



from collections.abc import Iterable

from google.protobuf.message import Message

from .common_pb2 import AvailableAbility, Point2D

class RequestQuery(Message):
    pathing: Iterable[RequestQueryPathing]
    abilities: Iterable[RequestQueryAvailableAbilities]
    placements: Iterable[RequestQueryBuildingPlacement]
    ignore_resource_requirements: bool
    def __init__(
        self,
        pathing: Iterable[RequestQueryPathing] = ...,
        abilities: Iterable[RequestQueryAvailableAbilities] = ...,
        placements: Iterable[RequestQueryBuildingPlacement] = ...,
        ignore_resource_requirements: bool = ...,
    ) -> None: ...

class ResponseQuery(Message):
    pathing: Iterable[ResponseQueryPathing]
    abilities: Iterable[ResponseQueryAvailableAbilities]
    placements: Iterable[ResponseQueryBuildingPlacement]
    def __init__(
        self,
        pathing: Iterable[ResponseQueryPathing] = ...,
        abilities: Iterable[ResponseQueryAvailableAbilities] = ...,
        placements: Iterable[ResponseQueryBuildingPlacement] = ...,
    ) -> None: ...

class RequestQueryPathing(Message):
    start_pos: Point2D
    unit_tag: int
    end_pos: Point2D
    def __init__(
        self,
        start_pos: Point2D = ...,
        unit_tag: int = ...,
        end_pos: Point2D = ...,
    ) -> None: ...

class ResponseQueryPathing(Message):
    distance: float
    def __init__(self, distance: float = ...) -> None: ...

class RequestQueryAvailableAbilities(Message):
    unit_tag: int
    def __init__(self, unit_tag: int = ...) -> None: ...

class ResponseQueryAvailableAbilities(Message):
    abilities: Iterable[AvailableAbility]
    unit_tag: int
    unit_type_id: int
    def __init__(
        self,
        abilities: Iterable[AvailableAbility] = ...,
        unit_tag: int = ...,
        unit_type_id: int = ...,
    ) -> None: ...

class RequestQueryBuildingPlacement(Message):
    ability_id: int
    target_pos: Point2D
    placing_unit_tag: int
    def __init__(
        self,
        ability_id: int = ...,
        target_pos: Point2D = ...,
        placing_unit_tag: int = ...,
    ) -> None: ...

class ResponseQueryBuildingPlacement(Message):
    result: int
    def __init__(self, result: int = ...) -> None: ...



# pyre-ignore-all-errors[14]
from __future__ import annotations

# DO NOT EDIT!
# This file was automatically generated by "generate_ids.py"

__all__ = ["unit_typeid", "ability_id", "upgrade_id", "buff_id", "effect_id"]



# pyre-ignore-all-errors[6, 11, 16]
"""
This class is very experimental and probably not up to date and needs to be refurbished.
If it works, you can watch replays with it.
"""

from __future__ import annotations

from sc2.bot_ai_internal import BotAIInternal
from sc2.data import Alert, Result
from sc2.ids.ability_id import AbilityId
from sc2.ids.upgrade_id import UpgradeId
from sc2.position import Point2
from sc2.unit import Unit
from sc2.units import Units


class ObserverAI(BotAIInternal):
    """Base class for bots."""

    @property
    def time(self) -> float:
        """Returns time in seconds, assumes the game is played on 'faster'"""
        return self.state.game_loop / 22.4  # / (1/1.4) * (1/16)

    @property
    def time_formatted(self) -> str:
        """Returns time as string in min:sec format"""
        t = self.time
        return f"{int(t // 60):02}:{int(t % 60):02}"

    def alert(self, alert_code: Alert) -> bool:
        """
        Check if alert is triggered in the current step.
        Possible alerts are listed here https://github.com/Blizzard/s2client-proto/blob/e38efed74c03bec90f74b330ea1adda9215e655f/s2clientprotocol/sc2api.proto#L679-L702

        Example use:

            from sc2.data import Alert
            if self.alert(Alert.AddOnComplete):
                print("Addon Complete")

        Alert codes::

            AlertError
            AddOnComplete
            BuildingComplete
            BuildingUnderAttack
            LarvaHatched
            MergeComplete
            MineralsExhausted
            MorphComplete
            MothershipComplete
            MULEExpired
            NuclearLaunchDetected
            NukeComplete
            NydusWormDetected
            ResearchComplete
            TrainError
            TrainUnitComplete
            TrainWorkerComplete
            TransformationComplete
            UnitUnderAttack
            UpgradeComplete
            VespeneExhausted
            WarpInComplete

        :param alert_code:
        """
        assert isinstance(alert_code, Alert), f"alert_code {alert_code} is no Alert"
        return alert_code.value in self.state.alerts

    @property
    def start_location(self) -> Point2:
        """
        Returns the spawn location of the bot, using the position of the first created townhall.
        This will be None if the bot is run on an arcade or custom map that does not feature townhalls at game start.
        """
        return self.game_info.player_start_location

    @property
    def enemy_start_locations(self) -> list[Point2]:
        """Possible start locations for enemies."""
        return self.game_info.start_locations

    async def get_available_abilities(
        self, units: list[Unit] | Units, ignore_resource_requirements: bool = False
    ) -> list[list[AbilityId]]:
        """Returns available abilities of one or more units. Right now only checks cooldown, energy cost, and whether the ability has been researched.

        Examples::

            units_abilities = await self.get_available_abilities(self.units)

        or::

            units_abilities = await self.get_available_abilities([self.units.random])

        :param units:
        :param ignore_resource_requirements:"""
        return await self.client.query_available_abilities(units, ignore_resource_requirements)

    async def on_unit_destroyed(self, unit_tag: int) -> None:
        """
        Override this in your bot class.
        This will event will be called when a unit (or structure, friendly or enemy) dies.
        For enemy units, this only works if the enemy unit was in vision on death.

        :param unit_tag:
        """

    async def on_unit_created(self, unit: Unit) -> None:
        """Override this in your bot class. This function is called when a unit is created.

        :param unit:"""

    async def on_building_construction_started(self, unit: Unit) -> None:
        """
        Override this in your bot class.
        This function is called when a building construction has started.

        :param unit:
        """

    async def on_building_construction_complete(self, unit: Unit) -> None:
        """
        Override this in your bot class. This function is called when a building
        construction is completed.

        :param unit:
        """

    async def on_upgrade_complete(self, upgrade: UpgradeId) -> None:
        """
        Override this in your bot class. This function is called with the upgrade id of an upgrade that was not finished last step and is now.

        :param upgrade:
        """

    async def on_start(self) -> None:
        """
        Override this in your bot class. This function is called after "on_start".
        At this point, game_data, game_info and the first iteration of game_state (self.state) are available.
        """

    async def on_step(self, iteration: int):
        """
        You need to implement this function!
        Override this in your bot class.
        This function is called on every game step (looped in realtime mode).

        :param iteration:
        """
        raise NotImplementedError

    async def on_end(self, game_result: Result) -> None:
        """Override this in your bot class. This function is called at the end of a game.

        :param game_result:"""



from pathlib import Path

from sc2.main import get_replay_version

THIS_FOLDER = Path(__file__).parent
REPLAY_PATHS = [path for path in (THIS_FOLDER / "replays").iterdir() if path.suffix == ".SC2Replay"]


def test_get_replay_version():
    for replay_path in REPLAY_PATHS:
        version = get_replay_version(replay_path)
        assert version == ("Base86383", "22EAC562CD0C6A31FB2C2C21E3AA3680")



from __future__ import annotations

import os
import platform
import re
import sys
from contextlib import suppress
from pathlib import Path

from loguru import logger

from sc2 import wsl

BASEDIR = {
    "Windows": "C:/Program Files (x86)/StarCraft II",
    "WSL1": "/mnt/c/Program Files (x86)/StarCraft II",
    "WSL2": "/mnt/c/Program Files (x86)/StarCraft II",
    "Darwin": "/Applications/StarCraft II",
    "Linux": "~/StarCraftII",
    "WineLinux": "~/.wine/drive_c/Program Files (x86)/StarCraft II",
}

USERPATH: dict[str, str | None] = {
    "Windows": "Documents\\StarCraft II\\ExecuteInfo.txt",
    "WSL1": "Documents/StarCraft II/ExecuteInfo.txt",
    "WSL2": "Documents/StarCraft II/ExecuteInfo.txt",
    "Darwin": "Library/Application Support/Blizzard/StarCraft II/ExecuteInfo.txt",
    "Linux": None,
    "WineLinux": None,
}

BINPATH = {
    "Windows": "SC2_x64.exe",
    "WSL1": "SC2_x64.exe",
    "WSL2": "SC2_x64.exe",
    "Darwin": "SC2.app/Contents/MacOS/SC2",
    "Linux": "SC2_x64",
    "WineLinux": "SC2_x64.exe",
}

CWD: dict[str, str | None] = {
    "Windows": "Support64",
    "WSL1": "Support64",
    "WSL2": "Support64",
    "Darwin": None,
    "Linux": None,
    "WineLinux": "Support64",
}


def platform_detect():
    pf = os.environ.get("SC2PF", platform.system())
    if pf == "Linux":
        return wsl.detect() or pf
    return pf


PF = platform_detect()


def get_home():
    """Get home directory of user, using Windows home directory for WSL."""
    if PF in {"WSL1", "WSL2"}:
        return wsl.get_wsl_home() or Path.home().expanduser()
    return Path.home().expanduser()


def get_user_sc2_install():
    """Attempts to find a user's SC2 install if their OS has ExecuteInfo.txt"""
    if USERPATH[PF]:
        einfo = str(get_home() / Path(USERPATH[PF]))
        if Path(einfo).is_file():
            with Path(einfo).open() as f:
                content = f.read()
            if content:
                base = re.search(r" = (.*)Versions", content).group(1)
                if PF in {"WSL1", "WSL2"}:
                    base = str(wsl.win_path_to_wsl_path(base))

                if Path(base).exists():
                    return base
    return None


def get_env() -> None:
    # TODO: Linux env conf from: https://github.com/deepmind/pysc2/blob/master/pysc2/run_configs/platforms.py
    return None


def get_runner_args(cwd):
    if "WINE" in os.environ:
        runner_file = Path(os.environ.get("WINE"))
        runner_file = runner_file if runner_file.is_file() else runner_file / "wine"
        """
        TODO Is converting linux path really necessary?
        That would convert
        '/home/burny/Games/battlenet/drive_c/Program Files (x86)/StarCraft II/Support64'
        to
        'Z:\\home\\burny\\Games\\battlenet\\drive_c\\Program Files (x86)\\StarCraft II\\Support64'
        """
        return [runner_file, "start", "/d", cwd, "/unix"]
    return []


def latest_executeble(versions_dir, base_build=None):
    latest = None

    if base_build is not None:
        with suppress(ValueError):
            latest = (
                int(base_build[4:]),
                max(p for p in versions_dir.iterdir() if p.is_dir() and p.name.startswith(str(base_build))),
            )

    if base_build is None or latest is None:
        latest = max((int(p.name[4:]), p) for p in versions_dir.iterdir() if p.is_dir() and p.name.startswith("Base"))

    version, path = latest

    if version < 55958:
        logger.critical("Your SC2 binary is too old. Upgrade to 3.16.1 or newer.")
        sys.exit(1)
    return path / BINPATH[PF]


class _MetaPaths(type):
    """ "Lazily loads paths to allow importing the library even if SC2 isn't installed."""

    def __setup(cls):
        if PF not in BASEDIR:
            logger.critical(f"Unsupported platform '{PF}'")
            sys.exit(1)

        try:
            base = os.environ.get("SC2PATH") or get_user_sc2_install() or BASEDIR[PF]
            cls.BASE = Path(base).expanduser()
            cls.EXECUTABLE = latest_executeble(cls.BASE / "Versions")
            cls.CWD = cls.BASE / CWD[PF] if CWD[PF] else None

            cls.REPLAYS = cls.BASE / "Replays"

            if (cls.BASE / "maps").exists():
                cls.MAPS = cls.BASE / "maps"
            else:
                cls.MAPS = cls.BASE / "Maps"
        except FileNotFoundError as e:
            logger.critical(f"SC2 installation not found: File '{e.filename}' does not exist.")
            sys.exit(1)

    def __getattr__(cls, attr):
        cls.__setup()
        return getattr(cls, attr)


class Paths(metaclass=_MetaPaths):
    """Paths for SC2 folders, lazily loaded using the above metaclass."""



from __future__ import annotations

from collections.abc import Iterable
from enum import Enum

from google.protobuf.message import Message

class ObservationUI(Message):
    groups: Iterable[ControlGroup]
    single: SinglePanel
    multi: MultiPanel
    cargo: CargoPanel
    production: ProductionPanel
    def __init__(
        self,
        groups: Iterable[ControlGroup] = ...,
        single: SinglePanel = ...,
        multi: MultiPanel = ...,
        cargo: CargoPanel = ...,
        production: ProductionPanel = ...,
    ) -> None: ...

class ControlGroup(Message):
    control_group_index: int
    leader_unit_type: int
    count: int
    def __init__(
        self,
        control_group_index: int = ...,
        leader_unit_type: int = ...,
        count: int = ...,
    ) -> None: ...

class UnitInfo(Message):
    unit_type: int
    player_relative: int
    health: int
    shields: int
    energy: int
    transport_slots_taken: int
    build_progress: float
    add_on: UnitInfo
    max_health: int
    max_shields: int
    max_energy: int
    def __init__(
        self,
        unit_type: int = ...,
        player_relative: int = ...,
        health: int = ...,
        shields: int = ...,
        energy: int = ...,
        transport_slots_taken: int = ...,
        build_progress: float = ...,
        add_on: UnitInfo = ...,
        max_health: int = ...,
        max_shields: int = ...,
        max_energy: int = ...,
    ) -> None: ...

class SinglePanel(Message):
    unit: UnitInfo
    attack_upgrade_level: int
    armor_upgrade_level: int
    shield_upgrade_level: int
    buffs: Iterable[int]
    def __init__(
        self,
        unit: UnitInfo = ...,
        attack_upgrade_level: int = ...,
        armor_upgrade_level: int = ...,
        shield_upgrade_level: int = ...,
        buffs: Iterable[int] = ...,
    ) -> None: ...

class MultiPanel(Message):
    units: Iterable[UnitInfo]
    def __init__(self, units: Iterable[UnitInfo] = ...) -> None: ...

class CargoPanel(Message):
    unit: UnitInfo
    passengers: Iterable[UnitInfo]
    slots_available: int
    def __init__(
        self,
        unit: UnitInfo = ...,
        passengers: Iterable[UnitInfo] = ...,
        slots_available: int = ...,
    ) -> None: ...

class BuildItem(Message):
    ability_id: int
    build_progress: float
    def __init__(self, ability_id: int = ..., build_progress: float = ...) -> None: ...

class ProductionPanel(Message):
    unit: UnitInfo
    build_queue: Iterable[UnitInfo]
    production_queue: Iterable[BuildItem]
    def __init__(
        self,
        unit: UnitInfo = ...,
        build_queue: Iterable[UnitInfo] = ...,
        production_queue: Iterable[BuildItem] = ...,
    ) -> None: ...

class ActionUI(Message):
    control_group: ActionControlGroup
    select_army: ActionSelectArmy
    select_warp_gates: ActionSelectWarpGates
    select_larva: ActionSelectLarva
    select_idle_worker: ActionSelectIdleWorker
    multi_panel: ActionMultiPanel
    cargo_panel: ActionCargoPanelUnload
    production_panel: ActionProductionPanelRemoveFromQueue
    toggle_autocast: ActionToggleAutocast
    def __init__(
        self,
        control_group: ActionControlGroup = ...,
        select_army: ActionSelectArmy = ...,
        select_warp_gates: ActionSelectWarpGates = ...,
        select_larva: ActionSelectLarva = ...,
        select_idle_worker: ActionSelectIdleWorker = ...,
        multi_panel: ActionMultiPanel = ...,
        cargo_panel: ActionCargoPanelUnload = ...,
        production_panel: ActionProductionPanelRemoveFromQueue = ...,
        toggle_autocast: ActionToggleAutocast = ...,
    ) -> None: ...

class ControlGroupAction(Enum):
    Recall: int
    Set: int
    Append: int
    SetAndSteal: int
    AppendAndSteal: int

class ActionControlGroup(Message):
    action: int
    control_group_index: int
    def __init__(self, action: int = ..., control_group_index: int = ...) -> None: ...

class ActionSelectArmy(Message):
    selection_add: bool
    def __init__(self, selection_add: bool = ...) -> None: ...

class ActionSelectWarpGates(Message):
    selection_add: bool
    def __init__(self, selection_add: bool = ...) -> None: ...

class ActionSelectLarva(Message):
    def __init__(self) -> None: ...

class ActionSelectIdleWorker(Message):
    class Type(Enum):
        Set: int
        Add: int
        All: int
        AddAll: int

    type: int
    def __init__(self, type: int = ...) -> None: ...

class ActionMultiPanel(Message):
    class Type(Enum):
        SingleSelect: int
        DeselectUnit: int
        SelectAllOfType: int
        DeselectAllOfType: int

    type: int
    unit_index: int
    def __init__(self, type: int = ..., unit_index: int = ...) -> None: ...

class ActionCargoPanelUnload(Message):
    unit_index: int
    def __init__(self, unit_index: int = ...) -> None: ...

class ActionProductionPanelRemoveFromQueue(Message):
    unit_index: int
    def __init__(self, unit_index: int = ...) -> None: ...

class ActionToggleAutocast(Message):
    ability_id: int
    def __init__(self, ability_id: int = ...) -> None: ...



"""
You can execute this test running the following command from the root python-sc2 folder:
uv run pytest test/test_pickled_data.py

This test/script uses the pickle files located in "python-sc2/test/pickle_data" generated by "generate_pickle_files_bot.py" file, which is a bot that starts a game on each of the maps defined in the main function.

It will load the pickle files, recreate the bot object from scratch and tests most of the bot properties and functions.
All functions that require some kind of query or interaction with the API directly will have to be tested in the "autotest_bot.py" in a live game.
"""

from __future__ import annotations

import lzma
import math
import pickle
import random
import sys
import unittest
from contextlib import suppress
from pathlib import Path
from typing import Any

from google.protobuf.internal import api_implementation
from hypothesis import given, settings
from hypothesis import strategies as st
from loguru import logger

from sc2.bot_ai import BotAI
from sc2.client import Client
from sc2.constants import ALL_GAS, CREATION_ABILITY_FIX
from sc2.data import CloakState, Race
from sc2.game_data import AbilityData, Cost, GameData
from sc2.game_info import GameInfo
from sc2.game_state import GameState
from sc2.ids.ability_id import AbilityId
from sc2.ids.buff_id import BuffId
from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.upgrade_id import UpgradeId
from sc2.pixel_map import PixelMap
from sc2.position import Point2, Point3, Rect, Size
from sc2.unit import Unit
from sc2.units import Units

MAPS: list[Path] = [
    map_path for map_path in (Path(__file__).parent / "pickle_data").iterdir() if map_path.suffix == ".xz"
]


def load_map_pickle_data(map_path: Path) -> tuple[Any, Any, Any]:
    with lzma.open(str(map_path.absolute()), "rb") as f:
        raw_game_data, raw_game_info, raw_observation = pickle.load(f)
        return raw_game_data, raw_game_info, raw_observation


def build_bot_object_from_pickle_data(raw_game_data, raw_game_info, raw_observation) -> BotAI:
    # Build fresh bot object, and load the pickled data into the bot object
    bot = BotAI()
    game_data = GameData(raw_game_data.data)
    game_info = GameInfo(raw_game_info.game_info)
    game_state = GameState(raw_observation)
    bot._initialize_variables()
    client = Client(True)
    bot._prepare_start(client=client, player_id=1, game_info=game_info, game_data=game_data)
    bot._prepare_step(state=game_state, proto_game_info=raw_game_info)
    return bot


def get_map_specific_bot(map_path: Path) -> BotAI:
    assert map_path in MAPS
    data = load_map_pickle_data(map_path)
    return build_bot_object_from_pickle_data(*data)


def test_protobuf_implementation():
    """Make sure that cpp is used as implementation"""
    # Doesn't seem to be implemented in newer python versions
    if sys.version_info < (3, 10) and sys.platform != "darwin":
        assert api_implementation.Type() == "cpp"


def test_bot_ai():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    # Test initial bot attributes at game start

    # Properties from _prepare_start
    assert 1 <= bot.player_id <= 2
    assert isinstance(bot.race, Race)
    assert isinstance(bot.enemy_race, Race)

    # Properties from _prepare_step
    assert bot.units.amount == bot.workers.amount
    assert bot.structures.amount == bot.townhalls.amount
    assert bot.workers.amount == 12
    assert bot.townhalls.amount == 1
    assert bot.gas_buildings.amount == 0
    assert bot.minerals == 50
    assert bot.vespene == 0
    assert bot.supply_army == 0
    assert bot.supply_workers == 12
    assert bot.supply_cap == 15
    assert bot.supply_used == 12
    assert bot.supply_left == 3
    assert bot.idle_worker_count == 0
    assert bot.army_count == 0

    # Test properties updated by "_prepare_units" function
    assert not bot.blips
    assert bot.units
    assert bot.structures
    assert not bot.enemy_units
    assert not bot.enemy_structures
    assert bot.mineral_field
    assert bot.vespene_geyser
    assert bot.resources
    assert len(bot.destructables) >= 0
    assert isinstance(bot.destructables, (list, set, dict))
    assert len(bot.watchtowers) >= 0
    assert bot.all_units
    assert bot.workers
    assert bot.townhalls
    assert not bot.gas_buildings

    # Test bot_ai functions
    assert bot.time == 0
    assert bot.time_formatted in {"0:00", "00:00"}
    assert bot.start_location is None  # Is populated by main.py
    bot.game_info.player_start_location = bot.townhalls.random.position
    assert bot.townhalls.random.position not in bot.enemy_start_locations
    assert bot.enemy_units == Units([], bot)
    assert bot.enemy_structures == Units([], bot)
    bot.game_info.map_ramps, bot.game_info.vision_blockers = bot.game_info._find_ramps_and_vision_blockers()
    assert bot.main_base_ramp  # Test if any ramp was found

    # The following functions need to be tested by autotest_bot.py because they use API query which isn't available here as this file only uses the pickle files and is not able to interact with the API as SC2 is not running while this test runs
    assert bot.can_feed(UnitTypeId.MARINE)
    assert bot.can_feed(UnitTypeId.SIEGETANK)
    assert not bot.can_feed(UnitTypeId.THOR)
    assert not bot.can_feed(UnitTypeId.BATTLECRUISER)
    assert not bot.can_feed(UnitTypeId.IMMORTAL)
    assert bot.can_afford(UnitTypeId.ZERGLING)
    assert bot.can_afford(UnitTypeId.MARINE)
    assert bot.can_afford(UnitTypeId.SCV)
    assert bot.can_afford(UnitTypeId.DRONE)
    assert bot.can_afford(UnitTypeId.PROBE)
    assert bot.can_afford(AbilityId.COMMANDCENTERTRAIN_SCV)
    assert bot.can_afford(UnitTypeId.MARINE)
    assert not bot.can_afford(UnitTypeId.SIEGETANK)
    assert not bot.can_afford(UnitTypeId.BATTLECRUISER)
    assert not bot.can_afford(UnitTypeId.MARAUDER)
    assert not bot.can_afford(UpgradeId.WARPGATERESEARCH)
    assert not bot.can_afford(AbilityId.RESEARCH_WARPGATE)

    # Store old values for minerals, vespene
    old_values = bot.minerals, bot.vespene, bot.supply_cap, bot.supply_left, bot.supply_used
    bot.vespene = 50
    assert bot.can_afford(UpgradeId.WARPGATERESEARCH)
    assert bot.can_afford(AbilityId.RESEARCH_WARPGATE)
    bot.minerals = 150
    bot.supply_cap = 15
    bot.supply_left = -1
    bot.supply_used = 16
    # Confirm that units that don't cost supply can be built while at negative supply using can_afford function
    assert bot.can_afford(UnitTypeId.GATEWAY)
    assert bot.can_afford(UnitTypeId.PYLON)
    assert bot.can_afford(UnitTypeId.OVERLORD)
    assert bot.can_afford(UnitTypeId.BANELING)
    assert not bot.can_afford(UnitTypeId.ZERGLING)
    assert not bot.can_afford(UnitTypeId.MARINE)
    bot.minerals, bot.vespene, bot.supply_cap, bot.supply_left, bot.supply_used = old_values

    worker = bot.workers.random
    assert bot.select_build_worker(worker.position) == worker
    for w in bot.workers:
        if w == worker:
            continue
        assert bot.select_build_worker(w.position) != worker
    assert bot.already_pending_upgrade(UpgradeId.STIMPACK) == 0
    assert bot.already_pending(UpgradeId.STIMPACK) == 0
    assert bot.already_pending(UnitTypeId.SCV) == 0
    assert bot.get_terrain_height(worker) > 0
    assert bot.in_placement_grid(worker)
    assert bot.in_pathing_grid(worker)
    # The pickle data was created by a terran bot, so there is no creep under any worker
    assert not bot.has_creep(worker)
    # Why did this stop working, not visible on first frame?
    assert bot.is_visible(worker), f"Visibility value at worker is {bot.state.visibility[worker.position.rounded]}"

    # Check price for morphing units and upgrades
    cost_100 = [
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL1,
        UpgradeId.TERRANSHIPWEAPONSLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEPLATINGLEVEL1,
        UpgradeId.TERRANVEHICLEARMORSLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL1,
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL1,
        UpgradeId.TERRANINFANTRYARMORSLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1,
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL1,
        UpgradeId.ZERGMELEEWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL1,
        UpgradeId.ZERGMISSILEWEAPONSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL1,
        UpgradeId.PROTOSSGROUNDARMORSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL1,
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1,
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1,
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1,
        UpgradeId.ZERGFLYERWEAPONSLEVEL1,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL1,
        UpgradeId.PROTOSSAIRWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1,
        UpgradeId.ZERGFLYERARMORSLEVEL1,
    ]
    cost_175 = [
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL2,
        UpgradeId.TERRANSHIPWEAPONSLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEPLATINGLEVEL2,
        UpgradeId.TERRANVEHICLEARMORSLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL2,
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
        UpgradeId.ZERGFLYERWEAPONSLEVEL2,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL2,
        UpgradeId.PROTOSSAIRWEAPONSLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
        UpgradeId.ZERGFLYERARMORSLEVEL2,
    ]
    cost_200 = [
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL2,
        UpgradeId.PROTOSSSHIELDSLEVEL2,
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL2,
        UpgradeId.ZERGGROUNDARMORSLEVEL2,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL3,
        UpgradeId.ZERGMELEEWEAPONSLEVEL3,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL3,
        UpgradeId.ZERGMISSILEWEAPONSLEVEL3,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL3,
        UpgradeId.PROTOSSGROUNDARMORSLEVEL3,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL3,
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL3,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL3,
        UpgradeId.TERRANINFANTRYARMORSLEVEL3,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL3,
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL3,
    ]
    cost_250 = [
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL3,
        UpgradeId.TERRANSHIPWEAPONSLEVEL3,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEPLATINGLEVEL3,
        UpgradeId.TERRANVEHICLEARMORSLEVEL3,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL3,
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL3,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
        UpgradeId.ZERGFLYERWEAPONSLEVEL3,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL3,
        UpgradeId.PROTOSSAIRWEAPONSLEVEL3,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
        UpgradeId.ZERGFLYERARMORSLEVEL3,
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL3,
        UpgradeId.ZERGGROUNDARMORSLEVEL3,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL3,
        UpgradeId.PROTOSSSHIELDSLEVEL3,
    ]

    cost_150 = [
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL1,
        UpgradeId.ZERGGROUNDARMORSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1,
        UpgradeId.PROTOSSSHIELDSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1,
        UpgradeId.PROTOSSSHIELDSLEVEL1,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL2,
        UpgradeId.ZERGMELEEWEAPONSLEVEL2,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL2,
        UpgradeId.ZERGMISSILEWEAPONSLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL2,
        UpgradeId.PROTOSSGROUNDARMORSLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL2,
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL2,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL2,
        UpgradeId.TERRANINFANTRYARMORSLEVEL2,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL2,
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL2,
    ]
    cost_list = [100, 175, 200, 250, 150]

    def calc_cost(item_id) -> Cost:
        if isinstance(item_id, AbilityId):
            # pyre-ignore[16]
            return bot.game_data.calculate_ability_cost(item_id)
        elif isinstance(item_id, UpgradeId):
            return bot.game_data.upgrades[item_id.value].cost
        elif isinstance(item_id, UnitTypeId):
            creation_ability: AbilityId = bot.game_data.units[item_id.value].creation_ability.exact_id
            return bot.game_data.calculate_ability_cost(creation_ability)
        return Cost(0, 0)

    def assert_cost(item_id, real_cost: Cost):
        assert calc_cost(item_id) == real_cost, f"Cost of {item_id} should be {real_cost} but is {calc_cost(item_id)}"

    for items, cost in zip([cost_100, cost_175, cost_200, cost_250, cost_150], cost_list):
        real_cost2: Cost = Cost(cost, cost)
        for item in items:
            assert_cost(item, real_cost2)
            assert bot.calculate_cost(item) == real_cost2, (
                f"Cost of {item} should be {real_cost2} but is {calc_cost(item)}"
            )

    # Do not use the generic research abilities in the bot when testing if you can afford it as these are wrong
    assert_cost(AbilityId.RESEARCH_ZERGFLYERARMOR, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_ZERGFLYERATTACK, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_ZERGGROUNDARMOR, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_ZERGMELEEWEAPONS, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_ZERGMISSILEWEAPONS, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_TERRANINFANTRYARMOR, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_TERRANINFANTRYWEAPONS, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_PROTOSSGROUNDARMOR, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_PROTOSSGROUNDWEAPONS, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_PROTOSSSHIELDS, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_TERRANSHIPWEAPONS, Cost(0, 0))
    assert_cost(AbilityId.RESEARCH_TERRANVEHICLEWEAPONS, Cost(0, 0))

    # Somehow this is 0, returned by the API
    assert_cost(AbilityId.BUILD_REACTOR, Cost(0, 0))
    # UnitTypeId.REACTOR has no creation ability (None)
    # assert_cost(UnitTypeId.REACTOR, Cost(50, 50))

    assert_cost(AbilityId.BUILD_REACTOR_BARRACKS, Cost(50, 50))
    assert_cost(UnitTypeId.BARRACKSREACTOR, Cost(50, 50))
    assert_cost(AbilityId.UPGRADETOORBITAL_ORBITALCOMMAND, Cost(150, 0))
    assert_cost(UnitTypeId.ORBITALCOMMAND, Cost(150, 0))
    assert_cost(AbilityId.UPGRADETOORBITAL_ORBITALCOMMAND, Cost(150, 0))

    assert bot.calculate_unit_value(UnitTypeId.ORBITALCOMMAND) == Cost(550, 0)
    assert bot.calculate_unit_value(UnitTypeId.RAVAGER) == Cost(100, 100)
    assert bot.calculate_unit_value(UnitTypeId.ARCHON) == Cost(175, 275)
    assert bot.calculate_unit_value(UnitTypeId.ADEPTPHASESHIFT) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.AUTOTURRET) == Cost(100, 0)
    assert bot.calculate_unit_value(UnitTypeId.INFESTORTERRAN) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.INFESTORTERRANBURROWED) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.LARVA) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.EGG) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.LOCUSTMP) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.LOCUSTMPFLYING) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.BROODLING) == Cost(0, 0)
    # Other and effects
    assert bot.calculate_unit_value(UnitTypeId.KD8CHARGE) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.RAVAGERCORROSIVEBILEMISSILE) == Cost(0, 0)
    assert bot.calculate_unit_value(UnitTypeId.VIPERACGLUESCREENDUMMY) == Cost(0, 0)

    assert bot.calculate_cost(UnitTypeId.BROODLORD) == Cost(150, 150)
    assert bot.calculate_cost(UnitTypeId.RAVAGER) == Cost(25, 75)
    assert bot.calculate_cost(UnitTypeId.BANELING) == Cost(25, 25)
    assert bot.calculate_cost(UnitTypeId.ORBITALCOMMAND) == Cost(150, 0)
    assert bot.calculate_cost(AbilityId.UPGRADETOORBITAL_ORBITALCOMMAND) == Cost(150, 0)
    assert bot.calculate_cost(UnitTypeId.REACTOR) == Cost(50, 50)
    assert bot.calculate_cost(UnitTypeId.TECHLAB) == Cost(50, 25)
    assert bot.calculate_cost(UnitTypeId.QUEEN) == Cost(150, 0)
    assert bot.calculate_cost(UnitTypeId.HATCHERY) == Cost(300, 0)
    assert bot.calculate_cost(UnitTypeId.LAIR) == Cost(150, 100)
    assert bot.calculate_cost(UnitTypeId.HIVE) == Cost(200, 150)
    assert bot.calculate_cost(UnitTypeId.DRONE) == Cost(50, 0)
    assert bot.calculate_cost(UnitTypeId.SCV) == Cost(50, 0)
    assert bot.calculate_cost(UnitTypeId.PROBE) == Cost(50, 0)
    assert bot.calculate_cost(UnitTypeId.SPIRE) == Cost(200, 200)
    assert bot.calculate_cost(UnitTypeId.ARCHON) == bot.calculate_unit_value(UnitTypeId.ARCHON)

    assert_cost(AbilityId.MORPHTOBROODLORD_BROODLORD, Cost(150, 150))
    assert_cost(AbilityId.MORPHTORAVAGER_RAVAGER, Cost(25, 75))
    assert_cost(AbilityId.MORPH_LURKER, Cost(50, 100))
    assert_cost(AbilityId.MORPHZERGLINGTOBANELING_BANELING, Cost(25, 25))

    assert Cost(100, 50) == 2 * Cost(50, 25)
    assert Cost(100, 50) == Cost(50, 25) * 2
    assert Cost(50, 25) + Cost(50, 25) == Cost(50, 25) * 2
    assert Cost(50, 25) + Cost(50, 25) == 2 * Cost(50, 25)
    assert Cost(50, 25) != Cost(50, 25) * 2
    assert Cost(100, 50) - Cost(50, 25) == Cost(50, 25)

    assert bot.calculate_supply_cost(UnitTypeId.BARRACKS) == 0
    assert bot.calculate_supply_cost(UnitTypeId.HATCHERY) == 0
    assert bot.calculate_supply_cost(UnitTypeId.OVERLORD) == 0
    assert bot.calculate_supply_cost(UnitTypeId.ZERGLING) == 1
    assert bot.calculate_supply_cost(UnitTypeId.MARINE) == 1
    assert bot.calculate_supply_cost(UnitTypeId.BANELING) == 0
    assert bot.calculate_supply_cost(UnitTypeId.QUEEN) == 2
    assert bot.calculate_supply_cost(UnitTypeId.ROACH) == 2
    assert bot.calculate_supply_cost(UnitTypeId.RAVAGER) == 1
    assert bot.calculate_supply_cost(UnitTypeId.CORRUPTOR) == 2
    assert bot.calculate_supply_cost(UnitTypeId.BROODLORD) == 2
    assert bot.calculate_supply_cost(UnitTypeId.HYDRALISK) == 2
    assert bot.calculate_supply_cost(UnitTypeId.LURKERMP) == 1


def test_game_info():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    # Test if main base ramp works
    bot.game_info.map_ramps, bot.game_info.vision_blockers = bot.game_info._find_ramps_and_vision_blockers()
    game_info: GameInfo = bot.game_info

    bot.game_info.player_start_location = bot.townhalls.random.position

    # Test game info object
    assert len(game_info.players) == 2
    assert game_info.map_name
    assert game_info.local_map_path
    assert game_info.map_size
    assert game_info.pathing_grid
    assert game_info.terrain_height
    assert game_info.placement_grid
    assert game_info.playable_area
    assert game_info.map_center
    assert game_info.map_ramps
    assert game_info.player_races
    assert game_info.start_locations
    assert game_info.player_start_location


def test_game_data():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    game_data = bot.game_data

    assert game_data.abilities
    for ability_data in game_data.abilities.values():
        assert ability_data.id
        assert ability_data.exact_id
        assert ability_data.friendly_name
        # Doesnt work for all AbilityData (may return empty string or no cost)
        assert isinstance(ability_data.link_name, str)
        assert isinstance(ability_data.button_name, str)
        assert isinstance(ability_data.is_free_morph, bool)
        assert isinstance(ability_data.cost, Cost)

    assert game_data.units
    for unit_data in game_data.units.values():
        with suppress(ValueError):
            assert unit_data.id
        assert unit_data.name
        assert isinstance(unit_data.creation_ability, (AbilityData, type(None)))
        assert isinstance(unit_data.footprint_radius, (float, type(None)))
        # TODO Fails on newer python versions
        # assert isinstance(unit_data.attributes, RepeatedScalarContainer)
        assert isinstance(unit_data.has_minerals, bool)
        assert isinstance(unit_data.has_vespene, bool)
        assert isinstance(unit_data.cargo_size, int)
        assert isinstance(unit_data.tech_requirement, (UnitTypeId, type(None)))
        assert isinstance(unit_data.tech_alias, (list, type(None)))
        assert isinstance(unit_data.unit_alias, (UnitTypeId, type(None)))
        assert isinstance(unit_data.race, Race)
        assert isinstance(unit_data.cost_zerg_corrected, Cost)
        assert isinstance(unit_data.morph_cost, (Cost, type(None)))

    assert game_data.upgrades
    for upgrade_data in game_data.upgrades.values():
        assert isinstance(upgrade_data.name, str)
        assert isinstance(upgrade_data.research_ability, (AbilityData, type(None)))
        assert isinstance(upgrade_data.cost, Cost)


def test_game_state():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    state = bot.state

    assert not state.actions
    assert not state.action_errors
    assert not state.actions_unit_commands
    assert not state.actions_toggle_autocast
    assert not state.dead_units
    assert not state.alerts
    assert not state.player_result
    assert not state.chat
    assert state.common
    assert state.psionic_matrix
    assert state.game_loop == 0
    assert state.score
    assert not state.upgrades
    assert not state.dead_units
    assert state.visibility
    assert state.creep
    assert not state.effects


def test_pixelmap():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    pathing_grid: PixelMap = bot.game_info.pathing_grid
    assert pathing_grid.bits_per_pixel
    assert pathing_grid.bytes_per_pixel == pathing_grid.bits_per_pixel // 8
    assert not pathing_grid.is_set(Point2((0, 0)))
    assert pathing_grid.is_empty(Point2((0, 0)))
    pathing_grid[Point2((0, 0))] = 123
    assert pathing_grid.is_set(Point2((0, 0)))
    assert not pathing_grid.is_empty(Point2((0, 0)))
    pathing_grid.flood_fill_all(lambda i: True)
    pathing_grid.copy()
    pathing_grid.print()


def test_blip():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    # TODO this needs to be done in a test bot that has a sensor tower
    # blips are enemy dots on the minimap that are out of vision


def test_score():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    assert bot.state.score
    assert bot.state.score.summary


def test_unit():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    scv: Unit = bot.workers.random
    townhall: Unit = bot.townhalls.first

    assert scv.name
    assert scv.race
    assert scv.tag
    assert not scv.is_structure
    assert townhall.is_structure
    assert scv.is_light
    assert not townhall.is_light
    assert not scv.is_armored
    assert townhall.is_armored
    assert scv.is_biological
    assert not townhall.is_biological
    assert scv.is_mechanical
    assert townhall.is_mechanical
    assert not scv.is_massive
    assert not townhall.is_massive
    assert not scv.is_psionic
    assert not townhall.is_psionic
    assert scv.tech_alias is None
    assert townhall.tech_alias is None
    assert scv.unit_alias is None
    assert townhall.unit_alias is None
    assert scv.can_attack
    assert not townhall.can_attack
    assert not scv.can_attack_both
    assert not townhall.can_attack_both
    assert scv.can_attack_ground
    assert not townhall.can_attack_ground
    assert scv.ground_dps
    assert not townhall.ground_dps
    assert scv.ground_range
    assert not townhall.ground_range
    assert not scv.can_attack_air
    assert not townhall.can_attack_air
    assert not scv.air_dps
    assert not townhall.air_dps
    assert not scv.air_range
    assert not townhall.air_range
    assert not scv.bonus_damage
    assert not townhall.bonus_damage
    assert not scv.armor
    assert townhall.armor
    assert scv.sight_range
    assert townhall.sight_range
    assert scv.movement_speed
    assert scv.real_speed == scv.movement_speed
    assert not townhall.movement_speed
    assert townhall.real_speed == townhall.movement_speed
    assert abs(scv.distance_per_step - 0.502231) < 1e-3
    assert not townhall.distance_per_step
    assert scv.distance_to_weapon_ready == 0
    assert not townhall.distance_to_weapon_ready
    assert not scv.is_mineral_field
    assert not townhall.is_mineral_field
    assert not scv.is_vespene_geyser
    assert not townhall.is_vespene_geyser
    assert scv.health
    assert townhall.health
    assert scv.health_max
    assert townhall.health_max
    assert scv.health_percentage
    assert townhall.health_percentage
    assert not scv.shield
    assert not townhall.shield
    assert not scv.shield_max
    assert not townhall.shield_max
    assert not scv.shield_percentage
    assert not townhall.shield_percentage
    assert scv.shield_health_percentage == 1
    assert townhall.shield_health_percentage == 1
    assert not scv.energy
    assert not townhall.energy
    assert not scv.energy_max
    assert not townhall.energy_max
    assert not scv.energy_percentage
    assert not townhall.energy_percentage
    assert not scv.age_in_frames
    assert not townhall.age_in_frames
    assert not scv.age
    assert not townhall.age
    assert not scv.is_memory
    assert not townhall.is_memory
    assert not scv.is_snapshot
    assert not townhall.is_snapshot
    assert scv.is_visible
    assert townhall.is_visible
    assert not scv.is_placeholder
    assert not townhall.is_placeholder
    assert scv.alliance
    assert townhall.alliance
    assert scv.is_mine
    assert townhall.is_mine
    assert not scv.is_enemy
    assert not townhall.is_enemy
    assert scv.owner_id
    assert townhall.owner_id
    assert scv.position
    assert townhall.position
    assert scv.position3d
    assert townhall.position3d
    assert scv.distance_to(townhall)
    assert townhall.distance_to(scv)
    # assert scv.facing
    assert townhall.facing
    assert scv.radius
    assert townhall.radius
    assert scv.build_progress
    assert townhall.build_progress
    assert scv.is_ready
    assert townhall.is_ready
    assert scv.cloak == CloakState.NotCloaked
    assert townhall.cloak == CloakState.NotCloaked
    assert not scv.is_cloaked
    assert not townhall.is_cloaked
    assert not scv.is_revealed
    assert not townhall.is_revealed
    assert scv.can_be_attacked
    assert townhall.can_be_attacked
    assert not scv.buffs
    assert not townhall.buffs
    assert not scv.is_carrying_minerals
    assert not townhall.is_carrying_minerals
    assert not scv.is_carrying_vespene
    assert not townhall.is_carrying_vespene
    assert not scv.is_carrying_resource
    assert not townhall.is_carrying_resource
    assert not scv.detect_range
    assert not townhall.detect_range
    assert not scv.radar_range
    assert not townhall.radar_range
    assert not scv.is_selected
    assert not townhall.is_selected
    assert scv.is_on_screen
    assert townhall.is_on_screen
    assert not scv.is_blip
    assert not townhall.is_blip
    assert not scv.is_powered
    assert not townhall.is_powered
    assert scv.is_active
    assert not townhall.is_active
    assert not scv.mineral_contents
    assert not townhall.mineral_contents
    assert not scv.vespene_contents
    assert not townhall.vespene_contents
    assert not scv.has_vespene
    assert not townhall.has_vespene
    assert not scv.is_flying
    assert not townhall.is_flying
    assert not scv.is_burrowed
    assert not townhall.is_burrowed
    assert not scv.is_hallucination
    assert not townhall.is_hallucination
    assert not scv.buff_duration_remain
    assert not townhall.buff_duration_remain
    assert not scv.buff_duration_max
    assert not townhall.buff_duration_max
    assert scv.orders
    assert not townhall.orders
    assert scv.order_target
    assert not townhall.order_target
    assert not scv.is_idle
    assert townhall.is_idle
    assert not scv.is_using_ability(AbilityId.TERRANBUILD_SUPPLYDEPOT)
    assert not townhall.is_using_ability(AbilityId.COMMANDCENTERTRAIN_SCV)
    assert not scv.is_moving
    assert not townhall.is_moving
    assert not scv.is_attacking
    assert not townhall.is_attacking
    assert not scv.is_patrolling
    assert not townhall.is_patrolling
    assert scv.is_gathering
    assert not townhall.is_gathering
    assert not scv.is_returning
    assert not townhall.is_returning
    assert scv.is_collecting
    assert not townhall.is_collecting
    assert not scv.is_constructing_scv
    assert not townhall.is_constructing_scv
    assert not scv.is_transforming
    assert not townhall.is_transforming
    assert not scv.is_repairing
    assert not townhall.is_repairing
    assert not scv.add_on_tag
    assert not townhall.add_on_tag
    assert not scv.has_add_on
    assert not townhall.has_add_on
    assert not scv.has_techlab
    assert not townhall.has_techlab
    assert not scv.has_reactor
    assert not townhall.has_reactor
    assert scv.add_on_land_position
    assert townhall.add_on_land_position
    assert scv.add_on_position
    assert townhall.add_on_position
    assert not scv.passengers
    assert not townhall.passengers
    assert not scv.passengers_tags
    assert not townhall.passengers_tags
    assert not scv.cargo_used
    assert not townhall.cargo_used
    assert not scv.has_cargo
    assert not townhall.has_cargo
    assert scv.cargo_size
    assert not townhall.cargo_size
    assert not scv.cargo_max
    assert not townhall.cargo_max
    assert not scv.cargo_left
    assert not townhall.cargo_left
    assert not scv.assigned_harvesters
    assert townhall.assigned_harvesters == 12
    assert not scv.ideal_harvesters
    assert townhall.ideal_harvesters == 16
    assert not scv.surplus_harvesters
    assert townhall.surplus_harvesters == -4
    assert not scv.weapon_cooldown
    assert townhall.weapon_cooldown == -1
    assert scv.weapon_ready
    assert not townhall.weapon_ready
    assert not scv.engaged_target_tag
    assert not townhall.engaged_target_tag
    assert not scv.is_detector
    assert not townhall.is_detector
    assert scv.distance_to_squared(townhall)
    assert townhall.distance_to_squared(scv)
    assert scv.target_in_range(townhall, bonus_distance=5)
    assert not townhall.target_in_range(scv, bonus_distance=5)
    assert not scv.has_buff(BuffId.STIMPACK)
    assert not townhall.has_buff(BuffId.STIMPACK)

    assert scv.calculate_damage_vs_target(townhall)[0] == 4
    assert scv.calculate_damage_vs_target(townhall, ignore_armor=True)[0] == 5
    assert townhall.calculate_damage_vs_target(scv) == (0, 0, 0)
    assert townhall.calculate_damage_vs_target(scv, ignore_armor=True) == (0, 0, 0)

    # TODO create one of each unit in the pickle tests to do damage calculations without having to create a mock class for each unit

    assert scv.calculate_dps_vs_target(townhall) - 2.66 < 0.01
    assert scv.calculate_dps_vs_target(townhall, ignore_armor=True) - 3.33 < 0.01
    assert townhall.calculate_dps_vs_target(scv) == 0
    assert townhall.calculate_dps_vs_target(scv, ignore_armor=True) == 0

    assert scv.is_facing(townhall, angle_error=2 * math.pi)
    assert not scv.is_facing(townhall)
    assert townhall.is_facing(scv, angle_error=2 * math.pi)

    assert scv.footprint_radius == 0
    assert townhall.footprint_radius == 2.5

    # marauder1 = Unit(marauder_proto, bot)
    # marauder_15_hp = Unit(marauder_proto, bot)
    # marauder_15_hp._proto.health = 15
    # # Marauder1 should deal now 10+10vs_armored = 20 damage, but other marauder has 1 armor, so resulting damage should be 19
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp)[0] == 19
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp, ignore_armor=True)[0] == 20
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp, ignore_armor=True, include_overkill_damage=False)[0] == 15
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp, include_overkill_damage=False)[0] == 15
    #
    # marauder1._proto.attack_upgrade_level = 2
    # marauder_15_hp._proto.armor_upgrade_level = 1
    # # Marauder1 should deal now 12+12vs_armored = 24 damage, but other marauder has 2 armor, so resulting damage should be 22
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp)[0] == 22
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp, ignore_armor=True)[0] == 24
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp, ignore_armor=True, include_overkill_damage=False)[0] == 15
    # assert marauder1.calculate_damage_vs_target(marauder_15_hp, include_overkill_damage=False)[0] == 15


def test_units():
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))
    scvs = bot.workers
    townhalls = bot.townhalls

    assert scvs.amount
    assert townhalls.amount
    assert not scvs.empty
    assert not townhalls.empty
    assert scvs.exists
    assert townhalls.exists
    assert scvs.find_by_tag(scvs.random.tag)
    assert not townhalls.find_by_tag(0)
    assert scvs.first
    assert townhalls.first
    assert scvs.take(11)
    assert townhalls.take(1)
    assert scvs.random
    assert townhalls.random
    assert scvs.random_or(1)
    assert townhalls.random_or(0)
    assert scvs.random_group_of(11)
    assert not scvs.random_group_of(0)
    assert not townhalls.random_group_of(0)
    # assert not scvs.in_attack_range_of(townhalls.first)
    # assert not townhalls.in_attack_range_of(scvs.first)
    assert scvs.closest_distance_to(townhalls.first)
    assert scvs.closest_distance_to(townhalls.first.position)
    assert townhalls.closest_distance_to(scvs.first)
    assert scvs.furthest_distance_to(townhalls.first)
    assert scvs.furthest_distance_to(townhalls.first.position)
    assert townhalls.furthest_distance_to(scvs.first)
    assert scvs.closest_to(townhalls.first)
    assert scvs.closest_to(townhalls.first.position)
    assert townhalls.closest_to(scvs.first)
    assert scvs.furthest_to(townhalls.first)
    assert scvs.furthest_to(townhalls.first.position)
    assert townhalls.furthest_to(scvs.first)
    assert scvs.closer_than(10, townhalls.first)
    assert scvs.closer_than(10, townhalls.first.position)
    assert townhalls.closer_than(10, scvs.first)
    assert scvs.further_than(0, townhalls.first)
    assert scvs.further_than(0, townhalls.first.position)
    assert townhalls.further_than(0, scvs.first)
    assert townhalls.in_distance_between(scvs.first, 0, 999)
    assert townhalls.in_distance_between(scvs.first.position, 0, 999)
    assert townhalls.closest_n_units(scvs.first.position, n=1)
    assert townhalls.furthest_n_units(scvs.first.position, n=1)
    assert townhalls.in_distance_of_group(scvs, 999)
    assert townhalls.in_closest_distance_to_group(scvs)
    assert townhalls.n_closest_to_distance(scvs.first.position, 0, 1)
    assert townhalls.n_furthest_to_distance(scvs.first.position, 0, 1)

    empty_units = Units([], bot_object=bot)
    assert not empty_units
    assert not empty_units.closer_than(999, townhalls.first)
    assert not empty_units.further_than(0, townhalls.first)
    assert not empty_units.in_distance_between(townhalls.first, 0, 999)
    assert not empty_units.closest_n_units(townhalls.first, 0)
    assert not empty_units.furthest_n_units(townhalls.first, 0)

    assert scvs.subgroup(scvs)
    assert townhalls.subgroup(townhalls)
    assert scvs.filter(pred=lambda x: x.type_id == UnitTypeId.SCV)
    assert not townhalls.filter(pred=lambda x: x.type_id == UnitTypeId.NEXUS)
    assert scvs.sorted
    assert townhalls.sorted
    assert scvs.sorted_by_distance_to(townhalls.first)
    assert townhalls.sorted_by_distance_to(scvs.first)
    assert scvs.tags_in(scvs.tags)
    assert not townhalls.tags_in({0, 1, 2})
    assert not scvs.tags_not_in(scvs.tags)
    assert townhalls.tags_not_in({0, 1, 2})
    assert scvs.of_type(UnitTypeId.SCV)
    assert scvs.of_type([UnitTypeId.SCV])
    assert townhalls.of_type({UnitTypeId.COMMANDCENTER, UnitTypeId.COMMANDCENTERFLYING})
    assert not scvs.exclude_type(UnitTypeId.SCV)
    assert townhalls.exclude_type({UnitTypeId.COMMANDCENTERFLYING})
    assert not scvs.same_tech({UnitTypeId.PROBE})
    assert townhalls.same_tech({UnitTypeId.ORBITALCOMMAND})
    assert scvs.same_unit(UnitTypeId.SCV)
    assert townhalls.same_unit({UnitTypeId.COMMANDCENTERFLYING})
    assert scvs.center
    assert townhalls.center == townhalls.first.position
    assert not scvs.selected
    assert not townhalls.selected
    assert scvs.tags
    assert townhalls.tags
    assert scvs.ready
    assert townhalls.ready
    assert not scvs.not_ready
    assert not townhalls.not_ready
    assert not scvs.idle
    assert townhalls.idle
    assert scvs.owned
    assert townhalls.owned
    assert not scvs.enemy
    assert not townhalls.enemy
    assert not scvs.flying
    assert not townhalls.flying
    assert scvs.not_flying
    assert townhalls.not_flying
    assert not scvs.structure
    assert townhalls.structure
    assert scvs.not_structure
    assert not townhalls.not_structure
    assert scvs.gathering
    assert not townhalls.gathering
    assert not scvs.returning
    assert not townhalls.returning
    assert scvs.collecting
    assert not townhalls.collecting
    assert scvs.visible
    assert townhalls.visible
    assert not scvs.mineral_field
    assert not townhalls.mineral_field
    assert not scvs.vespene_geyser
    assert not townhalls.vespene_geyser
    assert scvs.prefer_idle
    assert townhalls.prefer_idle
    assert len(Unit.class_cache) == 2  # Filled with CC and SCV from previous tests
    assert len(scvs + townhalls) == 13
    assert hash(scvs + townhalls)
    assert scvs.copy()
    assert scvs.by_tag(scvs[0].tag)


def test_exact_creation_ability():
    try:
        from sc2.dicts.unit_abilities import UNIT_ABILITIES
        from sc2.dicts.unit_unit_alias import UNIT_UNIT_ALIAS
    except ImportError:
        logger.info("Import error: dict sc2/dicts/ are missing!")
        return
    test_case = unittest.TestCase()
    bot: BotAI = get_map_specific_bot(random.choice(MAPS))

    ignore_types = {
        UnitTypeId.ADEPTPHASESHIFT,
        UnitTypeId.ARBITERMP,
        UnitTypeId.BROODLING,
        UnitTypeId.BYPASSARMORDRONE,
        UnitTypeId.CORSAIRMP,
        UnitTypeId.EGG,
        UnitTypeId.ELSECARO_COLONIST_HUT,
        UnitTypeId.HERC,
        UnitTypeId.HERCPLACEMENT,
        UnitTypeId.INFESTEDTERRANSEGG,
        UnitTypeId.LARVA,
        UnitTypeId.NYDUSCANALCREEPER,
        UnitTypeId.QUEENMP,
        UnitTypeId.RAVENREPAIRDRONE,
        UnitTypeId.REPLICANT,
        UnitTypeId.SCOURGEMP,
        UnitTypeId.SCOUTMP,
        UnitTypeId.WARHOUND,
    }

    unit_types = list(UNIT_UNIT_ALIAS) + list(UNIT_UNIT_ALIAS.values()) + list(UNIT_ABILITIES) + list(ALL_GAS)
    unit_types_unique_sorted = sorted({t.name for t in unit_types})
    for unit_type_name in unit_types_unique_sorted:
        unit_type = UnitTypeId[unit_type_name]
        if unit_type in ignore_types:
            continue

        if unit_type in [
            UnitTypeId.ARCHON,
            UnitTypeId.ASSIMILATORRICH,
            UnitTypeId.EXTRACTORRICH,
            UnitTypeId.REFINERYRICH,
        ]:
            with test_case.assertRaises(AttributeError):
                _creation_ability = bot.game_data.units[unit_type.value].creation_ability.exact_id
            continue

        try:
            _creation_ability = bot.game_data.units[unit_type.value].creation_ability.exact_id
        except AttributeError:
            if unit_type not in CREATION_ABILITY_FIX:
                assert False, f"Unit type '{unit_type}' missing from CREATION_ABILITY_FIX"


def test_dicts():
    # May be missing but that should not fail the tests
    try:
        from sc2.dicts.unit_research_abilities import RESEARCH_INFO
    except ImportError:
        logger.info("Import error: dict sc2/dicts/unit_research_abilities.py is missing!")
        return

    bot: BotAI = get_map_specific_bot(random.choice(MAPS))

    unit_id: UnitTypeId
    data: dict
    for unit_id, data in RESEARCH_INFO.items():
        upgrade_id: UpgradeId
        for upgrade_id, upgrade_data in data.items():
            research_ability_correct: AbilityId = upgrade_data["ability"]
            research_ability_data_from_api = bot.game_data.upgrades[upgrade_id.value].research_ability
            if research_ability_data_from_api is None:
                continue
            research_ability_id_from_api: AbilityId = research_ability_data_from_api.exact_id
            if upgrade_id.value in {116, 117, 118}:
                # Research abilities for armory armor plating are mapped incorrectly in the API
                continue
            if research_ability_correct.value in {807, 1284}:
                # Test broke on windows
                continue
            assert research_ability_correct == research_ability_id_from_api, (
                f"Research abilities do not match: Correct one is {research_ability_correct} but API returned {research_ability_id_from_api}"
            )


@given(
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
)
@settings(max_examples=500)
def test_position_pointlike(x1, y1, x2, y2, x3, y3):
    pos1 = Point2((x1, y1))
    pos2 = Point2((x2, y2))
    pos3 = Point2((x3, y3))
    epsilon = 1e-3
    assert pos1.position == pos1
    dist = ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5
    assert abs(pos1.distance_to(pos2) - dist) <= epsilon
    assert abs(pos1.distance_to_point2(pos2) - dist) <= epsilon
    assert abs(pos1._distance_squared(pos2) ** 0.5 - dist) <= epsilon

    points = {pos2, pos3}
    points2 = {pos1, pos2, pos3}
    # All 3 points need to be different
    if len(points2) == 3:
        assert pos1.sort_by_distance(points2) == sorted(points2, key=lambda p: pos1._distance_squared(p))
        assert pos1.closest(points2) == pos1
        closest_point = min(points, key=lambda p: p._distance_squared(pos1))
        dist_closest_point = pos1._distance_squared(closest_point) ** 0.5
        furthest_point = max(points, key=lambda p: p._distance_squared(pos1))
        dist_furthest_point = pos1._distance_squared(furthest_point) ** 0.5

        # Distances between pos1-pos2 and pos1-pos3 might be the same, so the sorting might still be different, that's why I use a set here
        assert pos1.closest(points) in {p for p in points2 if abs(pos1.distance_to(p) - dist_closest_point) < epsilon}
        assert abs(pos1.distance_to_closest(points) - pos1._distance_squared(closest_point) ** 0.5) < epsilon
        assert pos1.furthest(points) in {p for p in points2 if abs(pos1.distance_to(p) - dist_furthest_point) < epsilon}
        assert abs(pos1.distance_to_furthest(points) - pos1._distance_squared(furthest_point) ** 0.5) < epsilon
        assert pos1.offset(pos2) == Point2((pos1.x + pos2.x, pos1.y + pos2.y))
        if pos1 != pos2:
            assert pos1.unit_axes_towards(pos2) != Point2((0, 0))

        if x3 > 0:
            temp_pos = pos1.towards(pos2, x3)
            if x3 <= pos1.distance_to(pos2):
                # Using "towards" function to go between pos1 and pos2
                dist1 = pos1.distance_to(temp_pos) + pos2.distance_to(temp_pos)
                dist2 = pos1.distance_to(pos2)
                assert abs(dist1 - dist2) <= epsilon
            else:
                # Using "towards" function to go past pos2
                dist1 = pos1.distance_to(pos2) + pos2.distance_to(temp_pos)
                dist2 = pos1.distance_to(temp_pos)
                assert abs(dist1 - dist2) <= epsilon
        elif x3 < 0:
            # Using "towards" function with a negative value
            temp_pos = pos1.towards(pos2, x3)
            dist1 = temp_pos.distance_to(pos1) + pos1.distance_to(pos2)
            dist2 = pos2.distance_to(temp_pos)
            assert abs(dist1 - dist2) <= epsilon

    assert pos1 == pos1
    assert pos2 == pos2
    assert pos3 == pos3
    assert isinstance(hash(pos1), int)
    assert isinstance(hash(pos2), int)
    assert isinstance(hash(pos3), int)


@given(
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
)
@settings(max_examples=500)
def test_position_point2(x1, y1, x2, y2):
    pos1 = Point2((x1, y1))
    pos2 = Point2((x2, y2))
    assert pos1.x == x1
    assert pos1.y == y1
    assert pos1.to2 == pos1
    assert pos1.to3 == Point3((x1, y1, 0))

    length1 = (pos1.x**2 + pos1.y**2) ** 0.5
    assert abs(pos1.length - length1) < 0.001
    if length1:
        normalized1 = pos1 / length1
        assert abs(pos1.normalized.is_same_as(pos1 / length1))
        assert abs(normalized1.length - 1) < 0.001
    length2 = (pos2.x**2 + pos2.y**2) ** 0.5
    assert abs(pos2.length - length2) < 0.001
    if length2:
        normalized2 = pos2 / length2
        assert abs(pos2.normalized.is_same_as(normalized2))
        assert abs(normalized2.length - 1) < 0.001

    assert isinstance(pos1.distance_to(pos2), float)
    assert isinstance(pos1.distance_to_point2(pos2), float)
    if x2 > 0:
        assert pos1.random_on_distance(x2) != pos1
        assert pos1.towards_with_random_angle(pos2, x2) != pos1
    assert pos1.towards_with_random_angle(pos2) != pos1
    if pos1 != pos2:
        dist = pos1.distance_to(pos2)
        intersections1 = pos1.circle_intersection(pos2, r=dist / 2)
        assert len(intersections1) == 1
        intersections2 = pos1.circle_intersection(pos2, r=dist * 2 / 3)
        assert len(intersections2) == 2
    neighbors4 = pos1.neighbors4
    assert len(neighbors4) == 4
    neighbors8 = pos1.neighbors8
    assert len(neighbors8) == 8

    assert pos1 + pos2 == Point2((x1 + x2, y1 + y2))
    assert pos1 - pos2 == Point2((x1 - x2, y1 - y2))
    assert pos1 * pos2 == Point2((x1 * x2, y1 * y2))
    if 0 not in {x2, y2}:
        assert pos2
        assert pos1 / pos2 == Point2((x1 / x2, y1 / y2))

    if pos1._distance_squared(pos2) < 0.1:
        assert pos1.is_same_as(pos2, dist=0.1)

    assert pos1.unit_axes_towards(pos2) == pos1.direction_vector(pos2)


@given(
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
)
@settings(max_examples=10)
def test_position_point3(x1, y1, z1):
    pos1 = Point3((x1, y1, z1))
    assert pos1.z == z1
    assert pos1.to3 == pos1


@given(st.integers(min_value=-1e5, max_value=1e5), st.integers(min_value=-1e5, max_value=1e5))
@settings(max_examples=20)
def test_position_size(w, h):
    size = Size((w, h))
    assert size.width == w
    assert size.height == h


@given(
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
    st.integers(min_value=-1e5, max_value=1e5),
)
@settings(max_examples=20)
def test_position_rect(x, y, w, h):
    rect = Rect((x, y, w, h))
    assert rect.x == x
    assert rect.y == y
    assert rect.width == w
    assert rect.height == h
    assert rect.right == x + w
    assert rect.top == y + h
    assert rect.size == Size((w, h))
    assert rect.center == Point2((rect.x + rect.width / 2, rect.y + rect.height / 2))
    assert rect.offset((1, 1)) == Rect((x + 1, y + 1, w, h))


def test_missing_enum():
    enum_number = 123456789
    enum_converted = BuffId(enum_number)
    assert enum_converted == BuffId.NULL


if __name__ == "__main__":
    test_unit()



from __future__ import annotations

from collections.abc import Iterable
from enum import Enum

from google.protobuf.message import Message

from .common_pb2 import ImageData, PointI, RectangleI

class ObservationFeatureLayer(Message):
    renders: FeatureLayers
    minimap_renders: FeatureLayersMinimap
    def __init__(
        self,
        renders: FeatureLayers = ...,
        minimap_renders: FeatureLayersMinimap = ...,
    ) -> None: ...

class FeatureLayers(Message):
    height_map: ImageData
    visibility_map: ImageData
    creep: ImageData
    power: ImageData
    player_id: ImageData
    unit_type: ImageData
    selected: ImageData
    unit_hit_points: ImageData
    unit_hit_points_ratio: ImageData
    unit_energy: ImageData
    unit_energy_ratio: ImageData
    unit_shields: ImageData
    unit_shields_ratio: ImageData
    player_relative: ImageData
    unit_density_aa: ImageData
    unit_density: ImageData
    effects: ImageData
    hallucinations: ImageData
    cloaked: ImageData
    blip: ImageData
    buffs: ImageData
    buff_duration: ImageData
    active: ImageData
    build_progress: ImageData
    buildable: ImageData
    pathable: ImageData
    placeholder: ImageData
    def __init__(
        self,
        height_map: ImageData = ...,
        visibility_map: ImageData = ...,
        creep: ImageData = ...,
        power: ImageData = ...,
        player_id: ImageData = ...,
        unit_type: ImageData = ...,
        selected: ImageData = ...,
        unit_hit_points: ImageData = ...,
        unit_hit_points_ratio: ImageData = ...,
        unit_energy: ImageData = ...,
        unit_energy_ratio: ImageData = ...,
        unit_shields: ImageData = ...,
        unit_shields_ratio: ImageData = ...,
        player_relative: ImageData = ...,
        unit_density_aa: ImageData = ...,
        unit_density: ImageData = ...,
        effects: ImageData = ...,
        hallucinations: ImageData = ...,
        cloaked: ImageData = ...,
        blip: ImageData = ...,
        buffs: ImageData = ...,
        buff_duration: ImageData = ...,
        active: ImageData = ...,
        build_progress: ImageData = ...,
        buildable: ImageData = ...,
        pathable: ImageData = ...,
        placeholder: ImageData = ...,
    ) -> None: ...

class FeatureLayersMinimap(Message):
    height_map: ImageData
    visibility_map: ImageData
    creep: ImageData
    camera: ImageData
    player_id: ImageData
    player_relative: ImageData
    selected: ImageData
    alerts: ImageData
    buildable: ImageData
    pathable: ImageData
    unit_type: ImageData
    def __init__(
        self,
        height_map: ImageData = ...,
        visibility_map: ImageData = ...,
        creep: ImageData = ...,
        camera: ImageData = ...,
        player_id: ImageData = ...,
        player_relative: ImageData = ...,
        selected: ImageData = ...,
        alerts: ImageData = ...,
        buildable: ImageData = ...,
        pathable: ImageData = ...,
        unit_type: ImageData = ...,
    ) -> None: ...

class ObservationRender(Message):
    map: ImageData
    minimap: ImageData
    def __init__(self, map: ImageData = ..., minimap: ImageData = ...) -> None: ...

class ActionSpatial(Message):
    unit_command: ActionSpatialUnitCommand
    camera_move: ActionSpatialCameraMove
    unit_selection_point: ActionSpatialUnitSelectionPoint
    unit_selection_rect: ActionSpatialUnitSelectionRect
    def __init__(
        self,
        unit_command: ActionSpatialUnitCommand = ...,
        camera_move: ActionSpatialCameraMove = ...,
        unit_selection_point: ActionSpatialUnitSelectionPoint = ...,
        unit_selection_rect: ActionSpatialUnitSelectionRect = ...,
    ) -> None: ...

class ActionSpatialUnitCommand(Message):
    ability_id: int
    target_screen_coord: PointI
    target_minimap_coord: PointI
    queue_command: bool
    def __init__(
        self,
        ability_id: int = ...,
        target_screen_coord: PointI = ...,
        target_minimap_coord: PointI = ...,
        queue_command: bool = ...,
    ) -> None: ...

class ActionSpatialCameraMove(Message):
    center_minimap: PointI
    def __init__(self, center_minimap: PointI = ...) -> None: ...

class Type(Enum):
    Select: int
    Toggle: int
    AllType: int
    AddAllType: int

class ActionSpatialUnitSelectionPoint(Message):
    selection_screen_coord: PointI
    type: int
    def __init__(self, selection_screen_coord: PointI = ..., type: int = ...) -> None: ...

class ActionSpatialUnitSelectionRect(Message):
    selection_screen_coord: Iterable[RectangleI]
    selection_add: bool
    def __init__(self, selection_screen_coord: Iterable[RectangleI] = ..., selection_add: bool = ...) -> None: ...



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path

import numpy as np

from s2clientprotocol.common_pb2 import ImageData
from sc2.position import Point2


class PixelMap:
    def __init__(self, proto: ImageData, in_bits: bool = False) -> None:
        """
        :param proto:
        :param in_bits:
        """
        self._proto = proto
        # Used for copying pixelmaps
        self._in_bits: bool = in_bits

        assert self.width * self.height == (8 if in_bits else 1) * len(self._proto.data), (
            f"{self.width * self.height} {(8 if in_bits else 1) * len(self._proto.data)}"
        )
        buffer_data = np.frombuffer(self._proto.data, dtype=np.uint8)
        if in_bits:
            buffer_data = np.unpackbits(buffer_data)
        self.data_numpy = buffer_data.reshape(self._proto.size.y, self._proto.size.x)

    @property
    def width(self) -> int:
        return self._proto.size.x

    @property
    def height(self) -> int:
        return self._proto.size.y

    @property
    def bits_per_pixel(self) -> int:
        return self._proto.bits_per_pixel

    @property
    def bytes_per_pixel(self) -> int:
        return self._proto.bits_per_pixel // 8

    def __getitem__(self, pos: tuple[int, int]) -> int:
        """Example usage: is_pathable = self._game_info.pathing_grid[Point2((20, 20))] != 0"""
        assert 0 <= pos[0] < self.width, f"x is {pos[0]}, self.width is {self.width}"
        assert 0 <= pos[1] < self.height, f"y is {pos[1]}, self.height is {self.height}"
        return int(self.data_numpy[pos[1], pos[0]])

    def __setitem__(self, pos: tuple[int, int], value: int) -> None:
        """Example usage: self._game_info.pathing_grid[Point2((20, 20))] = 255"""
        assert 0 <= pos[0] < self.width, f"x is {pos[0]}, self.width is {self.width}"
        assert 0 <= pos[1] < self.height, f"y is {pos[1]}, self.height is {self.height}"
        assert 0 <= value <= 254 * self._in_bits + 1, (
            f"value is {value}, it should be between 0 and {254 * self._in_bits + 1}"
        )
        assert isinstance(value, int), f"value is of type {type(value)}, it should be an integer"
        self.data_numpy[pos[1], pos[0]] = value

    def is_set(self, p: tuple[int, int]) -> bool:
        return self[p] != 0

    def is_empty(self, p: tuple[int, int]) -> bool:
        return not self.is_set(p)

    def copy(self) -> PixelMap:
        return PixelMap(self._proto, in_bits=self._in_bits)

    def flood_fill(self, start_point: Point2, pred: Callable[[int], bool]) -> set[Point2]:
        nodes: set[Point2] = set()
        queue: list[Point2] = [start_point]

        while queue:
            x, y = queue.pop()

            if not (0 <= x < self.width and 0 <= y < self.height):
                continue

            if Point2((x, y)) in nodes:
                continue

            if pred(self[x, y]):
                nodes.add(Point2((x, y)))
                queue += [Point2((x + a, y + b)) for a in [-1, 0, 1] for b in [-1, 0, 1] if not (a == 0 and b == 0)]
        return nodes

    def flood_fill_all(self, pred: Callable[[int], bool]) -> set[frozenset[Point2]]:
        groups: set[frozenset[Point2]] = set()

        for x in range(self.width):
            for y in range(self.height):
                if any((x, y) in g for g in groups):
                    continue

                if pred(self[x, y]):
                    groups.add(frozenset(self.flood_fill(Point2((x, y)), pred)))

        return groups

    def print(self, wide: bool = False) -> None:
        for y in range(self.height):
            for x in range(self.width):
                print("#" if self.is_set((x, y)) else " ", end=(" " if wide else ""))
            print("")

    def save_image(self, filename: str | Path) -> None:
        data = [(0, 0, self[x, y]) for y in range(self.height) for x in range(self.width)]

        from PIL import Image

        im = Image.new("RGB", (self.width, self.height))
        im.putdata(data)
        im.save(filename)

    def plot(self) -> None:
        import matplotlib.pyplot as plt

        plt.imshow(self.data_numpy, origin="lower")
        plt.show()



from __future__ import annotations

from pathlib import Path

from loguru import logger

from sc2.paths import Paths


def get(name: str) -> Map:
    # Iterate through 2 folder depths
    for map_dir in (p for p in Paths.MAPS.iterdir()):
        if map_dir.is_dir():
            for map_file in (p for p in map_dir.iterdir()):
                if Map.matches_target_map_name(map_file, name):
                    return Map(map_file)
        elif Map.matches_target_map_name(map_dir, name):
            return Map(map_dir)

    raise KeyError(f"Map '{name}' was not found. Please put the map file in \"/StarCraft II/Maps/\".")


class Map:
    def __init__(self, path: Path) -> None:
        self.path = path

        if self.path.is_absolute():
            try:
                self.relative_path = self.path.relative_to(Paths.MAPS)
            except ValueError:  # path not relative to basedir
                logger.warning(f"Using absolute path: {self.path}")
                self.relative_path = self.path
        else:
            self.relative_path = self.path

    @property
    def name(self) -> str:
        return self.path.stem

    @property
    def data(self) -> bytes:
        with Path(self.path).open("rb") as f:
            return f.read()

    def __repr__(self) -> str:
        return f"Map({self.path})"

    @classmethod
    def is_map_file(cls, file: Path) -> bool:
        return file.is_file() and file.suffix == ".SC2Map"

    @classmethod
    def matches_target_map_name(cls, file: Path, name: str) -> bool:
        return cls.is_map_file(file) and file.stem == name



from contextlib import suppress

from sc2.expiring_dict import ExpiringDict


def test_class():
    class State:
        def __init__(self):
            self.game_loop = 0

    class BotAI:
        def __init__(self):
            self.state = State()

        def increment(self, value=1):
            self.state.game_loop += value

    test_dict = {"hello": "its me mario", "does_this_work": "yes it works", "another_test": "yep this one also worked"}

    bot = BotAI()
    test = ExpiringDict(bot, max_age_frames=10)

    for key, value in test_dict.items():
        test[key] = value
    bot.increment()

    # Test len
    assert len(test) == 3

    # Test contains method
    assert "hello" in test
    assert "doesnt_exist" not in test

    # Get item
    result = test["hello"]
    assert result == "its me mario"

    # Get item that doesnt exist
    with suppress(KeyError):
        result = test["doesnt_exist"]
    assert result == test["hello"]

    # Set new item
    test["setitem"] = "test"

    assert len(test) == 4

    # Test iteration
    for key, item in test.items():
        assert key in test
        assert test[key] == item, (key, item)
        assert test.get(key) == item
        assert test.get(key, with_age=True)[0] == item
        assert test.get(key, with_age=True)[1] in {0, 1}

    c = 0
    for _key in test:
        c += 1
    assert c == 4

    c = 0
    for value in test.values():
        c += 1
    assert c == 4

    # Update from another dict
    updater_dict = {"new_key": "my_new_value"}
    test.update(updater_dict)
    assert "does_this_work" in test
    assert "new_key" in test

    # Test pop method
    new_key = test.pop("new_key")
    assert new_key == "my_new_value"

    # Advance the frames by 10, this means all entries should now be invalid
    bot.increment(10)

    assert len(test) == 0

    for _key in test:
        assert False

    for _value in test.values():
        assert False

    for _key, _value in test.items():
        assert False

    assert "new_key" not in test
    assert "setitem" not in test
    # len doesn't work at the moment how it should - all items in the dict are expired, so len should return 0
    assert len(test) == 0, len(test)

    # Test repr and str function
    test["another_test"] = "yep this one also worked"
    test["setitem"] = "test"
    assert repr(test) == "ExpiringDict('another_test': ('yep this one also worked', 11), 'setitem': ('test', 11))"
    assert str(test) == "ExpiringDict('another_test': ('yep this one also worked', 11), 'setitem': ('test', 11))"


if __name__ == "__main__":
    test_class()



# pyre-ignore-all-errors[16, 19]
"""For the list of enums, see here

https://github.com/Blizzard/s2client-proto/tree/bff45dae1fc685e6acbaae084670afb7d1c0832c/s2clientprotocol
"""

from __future__ import annotations

import enum

from s2clientprotocol import common_pb2 as common_pb
from s2clientprotocol import data_pb2 as data_pb
from s2clientprotocol import error_pb2 as error_pb
from s2clientprotocol import raw_pb2 as raw_pb
from s2clientprotocol import sc2api_pb2 as sc_pb
from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId

CreateGameError = enum.Enum("CreateGameError", sc_pb.ResponseCreateGame.Error.items())

PlayerType = enum.Enum("PlayerType", sc_pb.PlayerType.items())
Difficulty = enum.Enum("Difficulty", sc_pb.Difficulty.items())
AIBuild = enum.Enum("AIBuild", sc_pb.AIBuild.items())
Status = enum.Enum("Status", sc_pb.Status.items())
Result = enum.Enum("Result", sc_pb.Result.items())
Alert = enum.Enum("Alert", sc_pb.Alert.items())
ChatChannel = enum.Enum("ChatChannel", sc_pb.ActionChat.Channel.items())

Race = enum.Enum("Race", common_pb.Race.items())

DisplayType = enum.Enum("DisplayType", raw_pb.DisplayType.items())
Alliance = enum.Enum("Alliance", raw_pb.Alliance.items())
CloakState = enum.Enum("CloakState", raw_pb.CloakState.items())

Attribute = enum.Enum("Attribute", data_pb.Attribute.items())
TargetType = enum.Enum("TargetType", data_pb.Weapon.TargetType.items())
Target = enum.Enum("Target", data_pb.AbilityData.Target.items())

ActionResult = enum.Enum("ActionResult", error_pb.ActionResult.items())

# pyre-ignore[11]
race_worker: dict[Race, UnitTypeId] = {
    Race.Protoss: UnitTypeId.PROBE,
    Race.Terran: UnitTypeId.SCV,
    Race.Zerg: UnitTypeId.DRONE,
}

race_townhalls: dict[Race, set[UnitTypeId]] = {
    Race.Protoss: {UnitTypeId.NEXUS},
    Race.Terran: {
        UnitTypeId.COMMANDCENTER,
        UnitTypeId.ORBITALCOMMAND,
        UnitTypeId.PLANETARYFORTRESS,
        UnitTypeId.COMMANDCENTERFLYING,
        UnitTypeId.ORBITALCOMMANDFLYING,
    },
    Race.Zerg: {UnitTypeId.HATCHERY, UnitTypeId.LAIR, UnitTypeId.HIVE},
    Race.Random: {
        # Protoss
        UnitTypeId.NEXUS,
        # Terran
        UnitTypeId.COMMANDCENTER,
        UnitTypeId.ORBITALCOMMAND,
        UnitTypeId.PLANETARYFORTRESS,
        UnitTypeId.COMMANDCENTERFLYING,
        UnitTypeId.ORBITALCOMMANDFLYING,
        # Zerg
        UnitTypeId.HATCHERY,
        UnitTypeId.LAIR,
        UnitTypeId.HIVE,
    },
}

warpgate_abilities: dict[AbilityId, AbilityId] = {
    AbilityId.GATEWAYTRAIN_ZEALOT: AbilityId.WARPGATETRAIN_ZEALOT,
    AbilityId.GATEWAYTRAIN_STALKER: AbilityId.WARPGATETRAIN_STALKER,
    AbilityId.GATEWAYTRAIN_HIGHTEMPLAR: AbilityId.WARPGATETRAIN_HIGHTEMPLAR,
    AbilityId.GATEWAYTRAIN_DARKTEMPLAR: AbilityId.WARPGATETRAIN_DARKTEMPLAR,
    AbilityId.GATEWAYTRAIN_SENTRY: AbilityId.WARPGATETRAIN_SENTRY,
    AbilityId.TRAIN_ADEPT: AbilityId.TRAINWARP_ADEPT,
}

race_gas: dict[Race, UnitTypeId] = {
    Race.Protoss: UnitTypeId.ASSIMILATOR,
    Race.Terran: UnitTypeId.REFINERY,
    Race.Zerg: UnitTypeId.EXTRACTOR,
}



from collections.abc import Iterable
from enum import Enum

from google.protobuf.message import Message

class Target(Enum):
    # NONE: int
    Point: int
    Unit: int
    PointOrUnit: int
    PointOrNone: int

class AbilityData(Message):
    ability_id: int
    link_name: str
    link_index: int
    button_name: str
    friendly_name: str
    hotkey: str
    remaps_to_ability_id: int
    available: bool
    target: int
    allow_minimap: bool
    allow_autocast: bool
    is_building: bool
    footprint_radius: float
    is_instant_placement: bool
    cast_range: float
    def __init__(
        self,
        ability_id: int = ...,
        link_name: str = ...,
        link_index: int = ...,
        button_name: str = ...,
        friendly_name: str = ...,
        hotkey: str = ...,
        remaps_to_ability_id: int = ...,
        available: bool = ...,
        target: int = ...,
        allow_minimap: bool = ...,
        allow_autocast: bool = ...,
        is_building: bool = ...,
        footprint_radius: float = ...,
        is_instant_placement: bool = ...,
        cast_range: float = ...,
    ) -> None: ...

class Attribute(Enum):
    Light: int
    Armored: int
    Biological: int
    Mechanical: int
    Robotic: int
    Psionic: int
    Massive: int
    Structure: int
    Hover: int
    Heroic: int
    Summoned: int

class DamageBonus(Message):
    attribute: int
    bonus: float
    def __init__(self, attribute: int = ..., bonus: float = ...) -> None: ...

class TargetType(Enum):
    Ground: int
    Air: int
    Any: int

class Weapon(Message):
    type: int
    damage: float
    damage_bonus: Iterable[DamageBonus]
    attacks: int
    range: float
    speed: float
    def __init__(
        self,
        type: int = ...,
        damage: float = ...,
        damage_bonus: Iterable[DamageBonus] = ...,
        attacks: int = ...,
        range: float = ...,
        speed: float = ...,
    ) -> None: ...

class UnitTypeData(Message):
    unit_id: int
    name: str
    available: bool
    cargo_size: int
    mineral_cost: int
    vespene_cost: int
    food_required: float
    food_provided: float
    ability_id: int
    race: int
    build_time: float
    has_vespene: bool
    has_minerals: bool
    sight_range: float
    tech_alias: Iterable[int]
    unit_alias: int
    tech_requirement: int
    require_attached: bool
    attributes: Iterable[int]
    movement_speed: float
    armor: float
    weapons: Iterable[Weapon]
    def __init__(
        self,
        unit_id: int = ...,
        name: str = ...,
        available: bool = ...,
        cargo_size: int = ...,
        mineral_cost: int = ...,
        vespene_cost: int = ...,
        food_required: float = ...,
        food_provided: float = ...,
        ability_id: int = ...,
        race: int = ...,
        build_time: float = ...,
        has_vespene: bool = ...,
        has_minerals: bool = ...,
        sight_range: float = ...,
        tech_alias: Iterable[int] = ...,
        unit_alias: int = ...,
        tech_requirement: int = ...,
        require_attached: bool = ...,
        attributes: Iterable[int] = ...,
        movement_speed: float = ...,
        armor: float = ...,
        weapons: Iterable[Weapon] = ...,
    ) -> None: ...

class UpgradeData(Message):
    upgrade_id: int
    name: str
    mineral_cost: int
    vespene_cost: int
    research_time: float
    ability_id: int
    def __init__(
        self,
        upgrade_id: int = ...,
        name: str = ...,
        mineral_cost: int = ...,
        vespene_cost: int = ...,
        research_time: float = ...,
        ability_id: int = ...,
    ) -> None: ...

class BuffData(Message):
    buff_id: int
    name: str
    def __init__(self, buff_id: int = ..., name: str = ...) -> None: ...

class EffectData(Message):
    effect_id: int
    name: str
    friendly_name: str
    radius: float
    def __init__(
        self,
        effect_id: int = ...,
        name: str = ...,
        friendly_name: str = ...,
        radius: float = ...,
    ) -> None: ...



from collections.abc import Iterable
from enum import Enum

from google.protobuf.message import Message

from .common_pb2 import ImageData, Point, Point2D, RectangleI, Size2DI

class StartRaw(Message):
    map_size: Size2DI
    pathing_grid: ImageData
    terrain_height: ImageData
    placement_grid: ImageData
    playable_area: RectangleI
    start_locations: Iterable[Point2D]
    def __init__(
        self,
        map_size: Size2DI = ...,
        pathing_grid: ImageData = ...,
        terrain_height: ImageData = ...,
        placement_grid: ImageData = ...,
        playable_area: RectangleI = ...,
        start_locations: Iterable[Point2D] = ...,
    ) -> None: ...

class ObservationRaw(Message):
    player: PlayerRaw
    units: Iterable[Unit]
    map_state: MapState
    event: Event
    effects: Iterable[Effect]
    radar: Iterable[RadarRing]
    def __init__(
        self,
        player: PlayerRaw = ...,
        units: Iterable[Unit] = ...,
        map_state: MapState = ...,
        event: Event = ...,
        effects: Iterable[Effect] = ...,
        radar: Iterable[RadarRing] = ...,
    ) -> None: ...

class RadarRing(Message):
    pos: Point
    radius: float
    def __init__(self, pos: Point = ..., radius: float = ...) -> None: ...

class PowerSource(Message):
    pos: Point
    radius: float
    tag: int
    def __init__(self, pos: Point = ..., radius: float = ..., tag: int = ...) -> None: ...

class PlayerRaw(Message):
    power_sources: Iterable[PowerSource]
    camera: Point
    upgrade_ids: Iterable[int]
    def __init__(
        self,
        power_sources: Iterable[PowerSource] = ...,
        camera: Point = ...,
        upgrade_ids: Iterable[int] = ...,
    ) -> None: ...

class UnitOrder(Message):
    ability_id: int
    target_world_space_pos: Point
    target_unit_tag: int
    progress: float
    def __init__(
        self,
        ability_id: int = ...,
        target_world_space_pos: Point = ...,
        target_unit_tag: int = ...,
        progress: float = ...,
    ) -> None: ...

class DisplayType(Enum):
    Visible: int
    Snapshot: int
    Hidden: int
    Placeholder: int

class Alliance(Enum):
    Self: int
    Ally: int
    Neutral: int
    Enemy: int

class CloakState(Enum):
    CloakedUnknown: int
    Cloaked: int
    CloakedDetected: int
    NotCloaked: int
    CloakedAllied: int

class PassengerUnit(Message):
    tag: int
    health: float
    health_max: float
    shield: float
    shield_max: float
    energy: float
    energy_max: float
    unit_type: int
    def __init__(
        self,
        tag: int = ...,
        health: float = ...,
        health_max: float = ...,
        shield: float = ...,
        shield_max: float = ...,
        energy: float = ...,
        energy_max: float = ...,
        unit_type: int = ...,
    ) -> None: ...

class RallyTarget(Message):
    point: Point
    tag: int
    def __init__(self, point: Point = ..., tag: int = ...) -> None: ...

class Unit(Message):
    display_type: int
    alliance: int
    tag: int
    unit_type: int
    owner: int
    pos: Point
    facing: float
    radius: float
    build_progress: float
    cloak: int
    buff_ids: Iterable[int]
    detect_range: float
    radar_range: float
    is_selected: bool
    is_on_screen: bool
    is_blip: bool
    is_powered: bool
    is_active: bool
    attack_upgrade_level: int
    armor_upgrade_level: int
    shield_upgrade_level: int
    health: float
    health_max: float
    shield: float
    shield_max: float
    energy: float
    energy_max: float
    mineral_contents: int
    vespene_contents: int
    is_flying: bool
    is_burrowed: bool
    is_hallucination: bool
    orders: Iterable[UnitOrder]
    add_on_tag: int
    passengers: Iterable[PassengerUnit]
    cargo_space_taken: int
    cargo_space_max: int
    assigned_harvesters: int
    ideal_harvesters: int
    weapon_cooldown: float
    engaged_target_tag: int
    buff_duration_remain: int
    buff_duration_max: int
    rally_targets: Iterable[RallyTarget]
    def __init__(
        self,
        display_type: int = ...,
        alliance: int = ...,
        tag: int = ...,
        unit_type: int = ...,
        owner: int = ...,
        pos: Point = ...,
        facing: float = ...,
        radius: float = ...,
        build_progress: float = ...,
        cloak: int = ...,
        buff_ids: Iterable[int] = ...,
        detect_range: float = ...,
        radar_range: float = ...,
        is_selected: bool = ...,
        is_on_screen: bool = ...,
        is_blip: bool = ...,
        is_powered: bool = ...,
        is_active: bool = ...,
        attack_upgrade_level: int = ...,
        armor_upgrade_level: int = ...,
        shield_upgrade_level: int = ...,
        health: float = ...,
        health_max: float = ...,
        shield: float = ...,
        shield_max: float = ...,
        energy: float = ...,
        energy_max: float = ...,
        mineral_contents: int = ...,
        vespene_contents: int = ...,
        is_flying: bool = ...,
        is_burrowed: bool = ...,
        is_hallucination: bool = ...,
        orders: Iterable[UnitOrder] = ...,
        add_on_tag: int = ...,
        passengers: Iterable[PassengerUnit] = ...,
        cargo_space_taken: int = ...,
        cargo_space_max: int = ...,
        assigned_harvesters: int = ...,
        ideal_harvesters: int = ...,
        weapon_cooldown: float = ...,
        engaged_target_tag: int = ...,
        buff_duration_remain: int = ...,
        buff_duration_max: int = ...,
        rally_targets: Iterable[RallyTarget] = ...,
    ) -> None: ...

class MapState(Message):
    visibility: ImageData
    creep: ImageData
    def __init__(self, visibility: ImageData = ..., creep: ImageData = ...) -> None: ...

class Event(Message):
    dead_units: Iterable[int]
    def __init__(self, dead_units: Iterable[int] = ...) -> None: ...

class Effect(Message):
    effect_id: int
    pos: Iterable[Point2D]
    alliance: int
    owner: int
    radius: float
    def __init__(
        self,
        effect_id: int = ...,
        pos: Iterable[Point2D] = ...,
        alliance: int = ...,
        owner: int = ...,
        radius: float = ...,
    ) -> None: ...

class ActionRaw(Message):
    unit_command: ActionRawUnitCommand
    camera_move: ActionRawCameraMove
    toggle_autocast: ActionRawToggleAutocast
    def __init__(
        self,
        unit_command: ActionRawUnitCommand = ...,
        camera_move: ActionRawCameraMove = ...,
        toggle_autocast: ActionRawToggleAutocast = ...,
    ) -> None: ...

class ActionRawUnitCommand(Message):
    ability_id: int
    target_world_space_pos: Point2D
    target_unit_tag: int
    unit_tags: Iterable[int]
    queue_command: bool
    def __init__(
        self,
        ability_id: int = ...,
        target_world_space_pos: Point2D = ...,
        target_unit_tag: int = ...,
        unit_tags: Iterable[int] = ...,
        queue_command: bool = ...,
    ) -> None: ...

class ActionRawCameraMove(Message):
    center_world_space: Point
    def __init__(self, center_world_space: Point = ...) -> None: ...

class ActionRawToggleAutocast(Message):
    ability_id: int
    unit_tags: Iterable[int]
    def __init__(self, ability_id: int = ..., unit_tags: Iterable[int] = ...) -> None: ...
